__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

import heapq

from nml import expression, generic, grfstrings
from nml.actions import actionF
from nml.ast import base_statement

townname_serial = 1


class TownNames(base_statement.BaseStatement):
    """
    'town_names' ast node.

    @ivar name: Name ID of the town_name.
    @type name: C{None}, L{Identifier}, or L{ConstantNumeric}

    @ivar id_number: Allocated ID number for this town_name action F node.
    @type id_number: C{None} or C{int}

    @ivar style_name: Name of the translated string containing the name of the style, if any.
    @type style_name: C{None} or L{String}

    @ivar actFs: Action F instance needed before this one.
    @type actFs: C{list} of L{ActionF}

    @ivar parts: Parts of the names.
    @type parts: C{list} of L{TownNamesPart}

    @ivar param_list: Stored parameter list.
    @type param_list: C{list} of (L{TownNamesPart} or L{TownNamesParam})
    """

    def __init__(self, name, param_list, pos):
        base_statement.BaseStatement.__init__(self, "town_names-block", pos, False, False)
        self.name = name
        self.param_list = param_list

        self.id_number = None
        self.style_name = None
        self.actFs = []
        self.parts = []

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Town name")
        if self.name is not None:
            generic.print_dbg(indentation + 2, "Name:")
            self.name.debug_print(indentation + 4)
        for param in self.param_list:
            param.debug_print(indentation + 2)

    def pre_process(self):
        self.actFs = []
        self.parts = []
        for param in self.param_list:
            if isinstance(param, TownNamesPart):
                actFs, part = param.make_actions()
                self.actFs.extend(actFs)
                self.parts.append(part)
            else:
                if param.key.value != "styles":
                    raise generic.ScriptError("Expected 'styles' keyword.", param.pos)
                if len(param.value.params) > 0:
                    raise generic.ScriptError("Parameters of the 'styles' were not expected.", param.pos)
                if self.style_name is not None:
                    raise generic.ScriptError("'styles' is already defined.", self.pos)
                self.style_name = param.value

        if len(self.parts) == 0:
            raise generic.ScriptError("Missing name parts in a town_names item.", self.pos)

        # 'name' is actually a number.
        # Allocate it now, before the self.prepare_output() call (to prevent names to grab it).
        if self.name is not None and not isinstance(self.name, expression.Identifier):
            value = self.name.reduce_constant()
            if not isinstance(value, expression.ConstantNumeric):
                raise generic.ScriptError("ID should be an integer number.", self.pos)

            self.id_number = value.value
            if self.id_number < 0 or self.id_number > 0x7F:
                raise generic.ScriptError("ID must be a number between 0 and 0x7f (inclusive)", self.pos)

            if self.id_number not in actionF.free_numbers:
                raise generic.ScriptError("town names ID 0x{:x} is already used.".format(self.id_number), self.pos)
            actionF.free_numbers.remove(self.id_number)

    def __str__(self):
        ret = "town_names"
        if self.name is not None:
            ret += "({})".format(self.name)
        ret += "{{\n{}}}\n".format("".join(str(x) for x in self.param_list))
        return ret

    def get_action_list(self):
        return self.actFs + [actionF.ActionF(self.name, self.id_number, self.style_name, self.parts, self.pos)]


class TownNamesPart:
    """
    A class containing a town name part.

    @ivar pieces: Pieces of the town name part.
    @type pieces: C{list} of (L{TownNamesEntryDefinition} or L{TownNamesEntryText})

    @ivar pos: Position information of the parts block.
    @type pos: L{Position}

    @ivar startbit: First bit to use for this part, if defined.
    @type startbit: C{int} or C{None}

    @ivar num_bits: Number of bits to use, if defined.
    @type num_bits: C{int} or C{None}
    """

    def __init__(self, pieces, pos):
        self.pos = pos
        self.pieces = pieces

        self.startbit = None
        self.num_bits = None

    def make_actions(self):
        """
        Construct new actionF instances to store all pieces of this part, if needed.

        @return: Action F that should be defined before, and the processed part.
        @rtype: C{list} of L{ActionF}, L{TownNamesPart}
        """
        new_pieces = []
        for piece in self.pieces:
            piece.pre_process()
            if piece.probability.value == 0:
                generic.print_warning(
                    generic.Warning.OPTIMISATION, "Dropping town name piece with 0 probability.", piece.pos
                )
            else:
                new_pieces.append(piece)

        self.pieces = new_pieces

        actFs = self.move_pieces()
        if len(self.pieces) == 0:
            raise generic.ScriptError("Expected names and/or town_name references in the part.", self.pos)
        if len(self.pieces) > 255:
            raise generic.ScriptError(
                "Too many values in a part, found {:d}, maximum is 255".format(len(self.pieces)), self.pos
            )

        return actFs, self

    def move_pieces(self):
        """
        Move pieces to new action F instances to make it fit, if needed.

        @return: Created action F instances.
        @rtype:  C{list} of L{ActionF}

        @note: Function may change L{pieces}.
        """
        global townname_serial

        if len(self.pieces) <= 255:
            return []  # Trivially correct.

        # There are too many pieces.
        number_action_f = (len(self.pieces) + 254) // 255
        pow2 = 1
        while pow2 < number_action_f:
            pow2 = pow2 * 2
        if pow2 < 255:
            number_action_f = pow2

        heap = []  # Heap of (summed probability, subset-of-pieces)
        i = 0
        while i < number_action_f:
            # Index 'i' is added to have a unique sorting when lists have equal total probabilities.
            heapq.heappush(heap, (0, i, []))
            i = i + 1

        finished_actions = []
        # Index 'idx' is added to have a unique sorting when pieces have equal probabilities.
        rev_pieces = sorted(((p.probability.value, idx, p) for idx, p in enumerate(self.pieces)), reverse=True)
        for prob, _idx, piece in rev_pieces:
            while True:
                sub = heapq.heappop(heap)
                if len(sub[2]) < 255:
                    break
                # If a subset already has the max number of parts, consider it finished.
                finished_actions.append(sub)

            sub[2].append(piece)
            sub = (sub[0] + prob, sub[1], sub[2])
            heapq.heappush(heap, sub)

        finished_actions.extend(heap)

        # To ensure the chances do not get messed up due to one part needing fewer bits for its
        # selection, all parts are forced to use the same number of bits.
        max_prob = max(sub[0] for sub in finished_actions)
        num_bits = 1
        while (1 << num_bits) < max_prob:
            num_bits = num_bits + 1

        # Assign to action F
        actFs = []
        for _prob, _idx, sub in finished_actions:
            actF_name = expression.Identifier("**townname #{:d}**".format(townname_serial), None)
            townname_serial = townname_serial + 1
            town_part = TownNamesPart(sub, self.pos)
            town_part.set_numbits(num_bits)
            actF = actionF.ActionF(actF_name, None, None, [town_part], self.pos)
            actFs.append(actF)

            # Remove pieces of 'sub' from self.pieces
            counts = len(self.pieces), len(sub)
            sub_set = set(sub)
            self.pieces = [piece for piece in self.pieces if piece not in sub_set]
            assert len(self.pieces) == counts[0] - counts[1]

            self.pieces.append(TownNamesEntryDefinition(actF_name, expression.ConstantNumeric(1), self.pos))

        # update self.parts
        return actFs

    def assign_bits(self, startbit):
        """
        Assign bits for this piece.

        @param startbit: First bit free for use.
        @type  startbit: C{int}

        @return: Number of bits needed for this piece.
        @rtype:  C{int}
        """
        assert len(self.pieces) <= 255
        total = sum(piece.probability.value for piece in self.pieces)

        if self.startbit is None or self.startbit < startbit:
            self.startbit = startbit
        if self.num_bits is None:
            n = 1
            while total > (1 << n):
                n = n + 1
            self.num_bits = n
        assert (1 << self.num_bits) >= total
        return self.num_bits

    def set_numbits(self, numbits):
        """
        Set the number of bits that this part should use.
        """
        assert self.num_bits is None
        self.num_bits = numbits

    def debug_print(self, indentation):
        total = sum(piece.probability.value for piece in self.pieces)
        generic.print_dbg(indentation, "Town names part (total {:d})".format(total))
        for piece in self.pieces:
            piece.debug_print(indentation + 2, total)

    def __str__(self):
        return "{{\n\t{}\n}}\n".format("\n\t".join(str(piece) for piece in self.pieces))

    def get_length(self):
        size = 3  # textcount, firstbit, bitcount bytes.
        size += sum(piece.get_length() for piece in self.pieces)
        return size

    def resolve_townname_id(self):
        """
        Resolve the reference numbers to previous C{town_names} blocks.

        @return: Set of referenced C{town_names} block numbers.
        """
        blocks = set()
        for piece in self.pieces:
            block = piece.resolve_townname_id()
            if block is not None:
                blocks.add(block)
        return blocks

    def write(self, file):
        file.print_bytex(len(self.pieces))
        file.print_bytex(self.startbit)
        file.print_bytex(self.num_bits)
        for piece in self.pieces:
            piece.write(file)
            file.newline()


class TownNamesParam:
    """
    Class containing a parameter of a town name.
    Currently, known key/values:
     - 'styles'  / string expression
    """

    def __init__(self, key, value, pos):
        self.key = key
        self.value = value
        self.pos = pos

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Town names param")
        generic.print_dbg(indentation + 2, "Key:")
        self.key.debug_print(indentation + 4)
        generic.print_dbg(indentation + 2, "Value:")
        self.value.debug_print(indentation + 4)

    def __str__(self):
        return "{}: {};\n".format(self.key, self.value)


class TownNamesEntryDefinition:
    """
    An entry in a part referring to a non-final town name, with a given probability.

    @ivar def_number: Name or number referring to a previous town_names node.
    @type def_number: L{Identifier} or L{ConstantNumeric}

    @ivar number: Actual ID to use.
    @type number: C{None} or C{int}

    @ivar probability: Probability of picking this reference.
    @type probability: C{ConstantNumeric}

    @ivar pos: Position information of the parts block.
    @type pos: L{Position}
    """

    def __init__(self, def_number, probability, pos):
        self.def_number = def_number
        self.number = None
        self.probability = probability
        self.pos = pos

    def pre_process(self):
        self.number = None
        if not isinstance(self.def_number, expression.Identifier):
            self.def_number = self.def_number.reduce_constant()
            if not isinstance(self.def_number, expression.ConstantNumeric):
                raise generic.ScriptError("Reference to other town name ID should be an integer number.", self.pos)
            if self.def_number.value < 0 or self.def_number.value > 0x7F:
                raise generic.ScriptError(
                    "Reference number out of range (must be between 0 and 0x7f inclusive).", self.pos
                )

        self.probability = self.probability.reduce_constant()
        if not isinstance(self.probability, expression.ConstantNumeric):
            raise generic.ScriptError("Probability should be an integer number.", self.pos)
        if self.probability.value < 0 or self.probability.value > 0x7F:
            raise generic.ScriptError("Probability out of range (must be between 0 and 0x7f inclusive).", self.pos)

    def debug_print(self, indentation, total):
        if isinstance(self.def_number, expression.Identifier):
            name_text = "name '" + self.def_number.value + "'"
        else:
            name_text = "number 0x{:x}".format(self.def_number.value)

        generic.print_dbg(
            indentation,
            "Insert town_name ID {} with probability {:d}/{:d}".format(name_text, self.probability.value, total),
        )

    def __str__(self):
        return "town_names({}, {:d}),".format(str(self.def_number), self.probability.value)

    def get_length(self):
        return 2

    def resolve_townname_id(self):
        """
        Resolve the reference number to a previous C{town_names} block.

        @return: Number of the referenced C{town_names} block.
        """
        if isinstance(self.def_number, expression.Identifier):
            self.number = actionF.named_numbers.get(self.def_number.value)
            if self.number is None:
                raise generic.ScriptError(
                    'Town names name "{}" is not defined or points to a next town_names node'.format(
                        self.def_number.value
                    ),
                    self.pos,
                )
        else:
            self.number = self.def_number.value
            if self.number not in actionF.numbered_numbers:
                raise generic.ScriptError(
                    'Town names number "{}" is not defined or points to a next town_names node'.format(self.number),
                    self.pos,
                )
        return self.number

    def write(self, file):
        file.print_bytex(self.probability.value | 0x80)
        file.print_bytex(self.number)


class TownNamesEntryText:
    """
    An entry in a part, a text-string with a given probability.

    @ivar pos: Position information of the parts block.
    @type pos: L{Position}
    """

    def __init__(self, id, text, probability, pos):
        self.id = id
        self.text = text
        self.probability = probability
        self.pos = pos

    def pre_process(self):
        if self.id.value != "text":
            raise generic.ScriptError("Expected 'text' prefix.", self.pos)

        if not isinstance(self.text, expression.StringLiteral):
            raise generic.ScriptError("Expected string literal for the name.", self.pos)

        self.probability = self.probability.reduce_constant()
        if not isinstance(self.probability, expression.ConstantNumeric):
            raise generic.ScriptError("Probability should be an integer number.", self.pos)
        if self.probability.value < 0 or self.probability.value > 0x7F:
            raise generic.ScriptError("Probability out of range (must be between 0 and 0x7f inclusive).", self.pos)

    def debug_print(self, indentation, total):
        generic.print_dbg(
            indentation, "Text {} with probability {:d}/{:d}".format(self.text.value, self.probability.value, total)
        )

    def __str__(self):
        return "text({}, {:d}),".format(self.text, self.probability.value)

    def get_length(self):
        return 1 + grfstrings.get_string_size(self.text.value)  # probability, text

    def resolve_townname_id(self):
        """
        Resolve the reference number to a previous C{town_names} block.

        @return: C{None}, as being the block number of a referenced previous C{town_names} block.
        """
        return None

    def write(self, file):
        file.print_bytex(self.probability.value)
        file.print_string(self.text.value, final_zero=True)



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml.editors import extract_tables

# Define parts of np++ xml file
string1 = """\
<NotepadPlus>
    <UserLang name="nml" ext="nml pnml">
        <Settings>
            <Global caseIgnored="no" />
            <TreatAsSymbol comment="no" commentLine="yes" />
            <Prefix words1="no" words2="no" words3="no" words4="no" />
        </Settings>
        <KeywordLists>
            <Keywords name="Delimiters">000000</Keywords>
            <Keywords name="Folder+">{</Keywords>
            <Keywords name="Folder-">}</Keywords>
            <Keywords name="Operators">( ) , : ; [ ]</Keywords>
            <Keywords name="Comment">1/* 2*/ 0//</Keywords>
            <!-- blocks, functions and units -->
            <Keywords name="Words1">"""

string2 = """\
</Keywords>
            <!-- properties and variables-->
            <Keywords name="Words2">"""

string3 = """\
</Keywords>
            <!-- features -->
            <Keywords name="Words3">"""

string4 = """\
</Keywords>
            <!-- callbacks and constants -->
            <Keywords name="Words4">"""

string5 = """\
</Keywords>
        </KeywordLists>
        <Styles>
            <WordsStyle name="DEFAULT" styleID="11" fgColor="000000" bgColor="FFFFFF" fontName="" fontStyle="0" />
            <WordsStyle name="FOLDEROPEN" styleID="12" fgColor="000080" bgColor="FFFFFF" fontName="" fontStyle="1" />
            <WordsStyle name="FOLDERCLOSE" styleID="13" fgColor="000080" bgColor="FFFFFF" fontName="" fontStyle="1" />
            <WordsStyle name="KEYWORD1" styleID="5" fgColor="0000FF" bgColor="FFFFFF" fontName="" fontStyle="0" />
            <WordsStyle name="KEYWORD2" styleID="6" fgColor="800000" bgColor="FFFFFF" fontName="" fontStyle="0" />
            <WordsStyle name="KEYWORD3" styleID="7" fgColor="D1802E" bgColor="FFFFFF" fontName="" fontStyle="1" />
            <WordsStyle name="KEYWORD4" styleID="8" fgColor="008040" bgColor="FFFFFF" fontName="" fontStyle="0" />
            <WordsStyle name="COMMENT" styleID="1" fgColor="FF9900" bgColor="FFFFFF" fontName="" fontStyle="0" />
            <WordsStyle name="COMMENT LINE" styleID="2" fgColor="FF9900" bgColor="FFFFFF" fontName="" fontStyle="0" />
            <WordsStyle name="NUMBER" styleID="4" fgColor="000000" bgColor="FFFFFF" fontName="" fontStyle="0" />
            <WordsStyle name="OPERATOR" styleID="10" fgColor="000080" bgColor="FFFFFF" fontName="" fontStyle="1" />
            <WordsStyle name="DELIMINER1" styleID="14" fgColor="000000" bgColor="FFFFFF" fontName="" fontStyle="0" />
            <WordsStyle name="DELIMINER2" styleID="15" fgColor="000000" bgColor="FFFFFF" fontName="" fontStyle="0" />
            <WordsStyle name="DELIMINER3" styleID="16" fgColor="000000" bgColor="FFFFFF" fontName="" fontStyle="0" />
        </Styles>
    </UserLang>
</NotepadPlus>
"""


# Build np++ xml file
def write_file(fname):
    with open(fname, "w") as file:
        file.write(string1)
        file.write(" ".join(extract_tables.block_names_table))
        file.write(string2)
        file.write(" ".join(extract_tables.variables_names_table))
        file.write(string3)
        file.write(" ".join(extract_tables.feature_names_table))
        file.write(string4)
        file.write(" ".join(extract_tables.callback_names_table))
        file.write(string5)


def run():
    write_file("nml_notepadpp.xml")



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic, nmlop

from .base_expression import ConstantFloat, ConstantNumeric, Expression
from .boolean import Boolean
from .string_literal import StringLiteral
from .variable import Variable


class BinOp(Expression):
    def __init__(self, op, expr1, expr2, pos=None):
        self.op = op
        self.expr1 = expr1 if isinstance(expr1, Expression) else ConstantNumeric(expr1)
        self.expr2 = expr2 if isinstance(expr2, Expression) else ConstantNumeric(expr2)
        pos = pos or self.expr1.pos or self.expr2.pos
        self.expr1.pos = self.expr1.pos or pos
        self.expr2.pos = self.expr2.pos or pos
        Expression.__init__(self, pos)

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Binary operator, op =", self.op.token)
        self.expr1.debug_print(indentation + 2)
        self.expr2.debug_print(indentation + 2)

    def __str__(self):
        if self.op == nmlop.SUB and isinstance(self.expr1, ConstantNumeric) and self.expr1.value == 0:
            return "-" + str(self.expr2)
        return self.op.to_string(self.expr1, self.expr2)

    def get_priority(self, expr):
        """
        Get the priority of an expression. For optimalization reason we prefer complexer
        expressions (= low priority) on the left hand side of this expression. The following
        priorities are used:
        -1: everything that doesn't fit in one of the other categories.
         0: Variables to be parsed in a varaction2
         1: Expressions that can be parsed via actionD, with the exception of constant numbers
         2: constant numbers

        @param expr: The expression to get the priority of.
        @type  expr: L{Expression}

        @return: The priority for the given expression.
        @rtype: C{int}
        """
        if isinstance(expr, Variable):
            return 0
        if isinstance(expr, ConstantNumeric):
            return 2
        if expr.supported_by_actionD(False):
            return 1
        return -1

    def reduce(self, id_dicts=None, unknown_id_fatal=True):
        # Reducing a BinOp expression is done in several phases:
        # - Reduce both subexpressions.
        # - If both subexpressions are constant, compute the result and return it.
        # - If the operator allows it and the second expression is more complex than
        #   the first one swap them.
        # - If the operation is a no-op, delete it.
        # - Variables (as used in action2var) can have some computations attached to
        #   them, do that if possible.
        # - Try to merge multiple additions/subtractions with constant numbers

        # - Reduce both subexpressions.
        expr1 = self.expr1.reduce(id_dicts)
        expr2 = self.expr2.reduce(id_dicts)

        # Make sure the combination of operands / operator is valid
        if self.op.validate_func is not None:
            self.op.validate_func(expr1, expr2, self.pos)

        # - If both subexpressions are constant, compute the result and return it.
        if (
            isinstance(expr1, ConstantNumeric)
            and isinstance(expr2, ConstantNumeric)
            and self.op.compiletime_func is not None
        ):
            return ConstantNumeric(self.op.compiletime_func(expr1.value, expr2.value), self.pos)

        if isinstance(expr1, StringLiteral) and isinstance(expr2, StringLiteral):
            assert self.op == nmlop.ADD
            return StringLiteral(expr1.value + expr2.value, expr1.pos)

        if (
            isinstance(expr1, (ConstantNumeric, ConstantFloat))
            and isinstance(expr2, (ConstantNumeric, ConstantFloat))
            and self.op.compiletime_func is not None
        ):
            return ConstantFloat(self.op.compiletime_func(expr1.value, expr2.value), self.pos)

        # - If the operator allows it and the second expression is more complex than
        #   the first one swap them.
        op = self.op
        if op.commutative or op in (nmlop.CMP_LT, nmlop.CMP_GT):
            prio1 = self.get_priority(expr1)
            prio2 = self.get_priority(expr2)
            if prio2 < prio1:
                expr1, expr2 = expr2, expr1
                if op == nmlop.CMP_LT:
                    op = nmlop.CMP_GT
                elif op == nmlop.CMP_GT:
                    op = nmlop.CMP_LT

        # - If the operation is a no-op, delete it.
        if op == nmlop.AND and isinstance(expr2, ConstantNumeric) and (expr2.value == -1 or expr2.value == 0xFFFFFFFF):
            return expr1

        if op in (nmlop.DIV, nmlop.DIVU, nmlop.MUL) and isinstance(expr2, ConstantNumeric) and expr2.value == 1:
            return expr1

        if (
            op
            in (
                nmlop.ADD,
                nmlop.SUB,
                nmlop.SHIFTU_LEFT,
                nmlop.SHIFT_LEFT,
                nmlop.SHIFTU_RIGHT,
                nmlop.SHIFT_RIGHT,
                nmlop.OR,
                nmlop.XOR,
                nmlop.ROT_RIGHT,
            )
            and isinstance(expr2, ConstantNumeric)
            and expr2.value == 0
        ):
            return expr1

        # - Variables (as used in action2var) can have some computations attached to
        #   them, do that if possible.
        if isinstance(expr1, Variable) and expr2.supported_by_actionD(False):
            # An action2 Variable has some special fields (mask, add, div and mod) that can be used
            # to perform some operations on the value. These operations are faster than a normal
            # advanced varaction2 operator so we try to use them whenever we can.
            if op == nmlop.AND and expr1.add is None:
                expr1.mask = nmlop.AND(expr1.mask, expr2, self.pos).reduce(id_dicts)
                return expr1
            if op == nmlop.ADD and expr1.div is None and expr1.mod is None:
                if expr1.add is None:
                    expr1.add = expr2
                else:
                    expr1.add = nmlop.ADD(expr1.add, expr2, self.pos).reduce(id_dicts)
                return expr1
            if op == nmlop.SUB and expr1.div is None and expr1.mod is None:
                if expr1.add is None:
                    expr1.add = ConstantNumeric(0)
                expr1.add = nmlop.SUB(expr1.add, expr2, self.pos).reduce(id_dicts)
                return expr1
            # The div and mod fields cannot be used at the same time. Also whenever either of those
            # two are used the add field has to be set, so we change it to zero when it's not yet set.
            if op == nmlop.DIV and expr1.div is None and expr1.mod is None:
                if expr1.add is None:
                    expr1.add = ConstantNumeric(0)
                expr1.div = expr2
                return expr1
            if op == nmlop.MOD and expr1.div is None and expr1.mod is None:
                if expr1.add is None:
                    expr1.add = ConstantNumeric(0)
                expr1.mod = expr2
                return expr1
            # Since we have a lot of nml-variables that are in fact only the high bits of a nfo
            # variable it can happen that we want to shift back the variable to the left.
            # Don't use any extra opcodes but just reduce the shift-right in that case.
            if (
                op == nmlop.SHIFT_LEFT
                and isinstance(expr2, ConstantNumeric)
                and expr1.add is None
                and expr2.value < expr1.shift.value
            ):
                expr1.shift.value -= expr2.value
                expr1.mask = nmlop.SHIFT_LEFT(expr1.mask, expr2).reduce()
                return expr1

            if (
                (
                    op == nmlop.SHIFTU_RIGHT
                    or (
                        op == nmlop.SHIFT_RIGHT
                        and isinstance(expr1.mask, ConstantNumeric)
                        and expr1.mask.uvalue < 0x80000000
                    )
                )
                and isinstance(expr2, ConstantNumeric)
                and expr2.value >= 0
                and expr1.shift.value >= 0
                and (expr1.shift.value + expr2.value) < 32
            ):
                expr1.shift.value += expr2.value
                expr1.mask = nmlop.SHIFTU_RIGHT(expr1.mask, expr2).reduce()
                return expr1

        # - Try to merge multiple additions/subtractions with constant numbers
        if (
            op in (nmlop.ADD, nmlop.SUB)
            and isinstance(expr2, ConstantNumeric)
            and isinstance(expr1, BinOp)
            and expr1.op in (nmlop.ADD, nmlop.SUB)
            and isinstance(expr1.expr2, ConstantNumeric)
        ):
            val = expr2.value if op == nmlop.ADD else -expr2.value
            if expr1.op == nmlop.ADD:
                return nmlop.ADD(expr1.expr1, (expr1.expr2.value + val), self.pos).reduce()
            if expr1.op == nmlop.SUB:
                return nmlop.SUB(expr1.expr1, (expr1.expr2.value - val), self.pos).reduce()

        if op == nmlop.OR and isinstance(expr1, Boolean) and isinstance(expr2, Boolean):
            return Boolean(nmlop.OR(expr1.expr, expr2.expr, self.pos)).reduce(id_dicts)

        return BinOp(op, expr1, expr2, self.pos)

    def supported_by_action2(self, raise_error):
        if not self.op.act2_supports:
            token = " '{}'".format(self.op.token) if self.op.token else ""
            if raise_error:
                raise generic.ScriptError("Operator{} not supported in a switch-block".format(token), self.pos)
            return False
        return self.expr1.supported_by_action2(raise_error) and self.expr2.supported_by_action2(raise_error)

    def supported_by_actionD(self, raise_error):
        if not self.op.actd_supports:
            if raise_error:
                if self.op == nmlop.STO_PERM:
                    raise generic.ScriptError("STORE_PERM is only available in switch-blocks.", self.pos)
                elif self.op == nmlop.STO_TMP:
                    raise generic.ScriptError("STORE_TEMP is only available in switch-blocks.", self.pos)

                # default case
                token = " '{}'".format(self.op.token) if self.op.token else ""
                raise generic.ScriptError("Operator{} not supported in parameter assignment".format(token), self.pos)
            return False
        return self.expr1.supported_by_actionD(raise_error) and self.expr2.supported_by_actionD(raise_error)

    def collect_references(self):
        return self.expr1.collect_references() + self.expr2.collect_references()

    def is_read_only(self):
        return self.expr1.is_read_only() and self.expr2.is_read_only()

    def is_boolean(self):
        if self.op in (nmlop.AND, nmlop.OR, nmlop.XOR):
            return self.expr1.is_boolean() and self.expr2.is_boolean()
        return self.op.returns_boolean

    def __eq__(self, other):
        return (
            other is not None
            and isinstance(other, BinOp)
            and self.op == other.op
            and self.expr1 == other.expr1
            and self.expr2 == other.expr2
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((self.op, self.expr1, self.expr2))



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic


class Type:
    """
    Enum-type class of the various value types possible in NML
    """

    INTEGER = 0
    FLOAT = 1
    STRING_LITERAL = 2
    FUNCTION_PTR = 3
    SPRITEGROUP_REF = 4


class Expression:
    """
    Superclass for all expression classes.

    @ivar pos: Position of the data in the original file.
    @type pos: :L{Position}
    """

    def __init__(self, pos):
        self.pos = pos

    def debug_print(self, indentation):
        """
        Print all data with explanation of what it is to standard output.

        @param indentation: Indent all printed lines with at least
            C{indentation} spaces.
        """
        raise NotImplementedError("debug_print must be implemented in expression-subclass {!r}".format(type(self)))

    def __str__(self):
        """
        Convert this expression to a string representing this expression in valid NML-code.

        @return: A string representation of this expression.
        """
        raise NotImplementedError("__str__ must be implemented in expression-subclass {!r}".format(type(self)))

    def reduce(self, id_dicts=None, unknown_id_fatal=True):
        """
        Reduce this expression to the simplest representation possible.

        @param id_dicts: A list with dicts that are used to map identifiers
            to another (often numeric) representation.
        @param unknown_id_fatal: Is encountering an unknown identifier somewhere
            in this expression a fatal error?

        @return: A deep copy of this expression simplified as much as possible.
        """
        raise NotImplementedError("reduce must be implemented in expression-subclass {!r}".format(type(self)))

    def reduce_constant(self, id_dicts=None):
        """
        Reduce this expression and make sure the result is a constant number.

        @param id_dicts: A list with dicts that are used to map identifiers
            to another (often numeric) representation.

        @return: A constant number that is the result of this expression.
        """
        expr = self.reduce(id_dicts)
        if not isinstance(expr, ConstantNumeric):
            raise generic.ConstError(self.pos)
        return expr

    def supported_by_action2(self, raise_error):
        """
        Check if this expression can be used inside a switch-block.

        @param raise_error: If true raise a scripterror instead of returning false.

        @return: True if this expression can be calculated by advanced varaction2.
        """
        if raise_error:
            raise generic.ScriptError("This expression is not supported in a switch-block or produce-block", self.pos)
        return False

    def supported_by_actionD(self, raise_error):
        """
        Check if this expression can be used inside a parameter-assignment.

        @param raise_error: If true raise a scripterror instead of returning false.

        @return: True if this expression can be calculated by actionD.
        """
        if raise_error:
            raise generic.ScriptError("This expression can not be assigned to a parameter", self.pos)
        return False

    def collect_references(self):
        """
        This function should collect all references to other nodes from this instance.

        @return: A collection containing all links to other nodes.
        @rtype: C{iterable} of L{SpriteGroupRef}
        """
        return []

    def is_read_only(self):
        """
        Check if this expression store values.

        @return: True if the expression doesn't store values.
        """
        return True

    def is_boolean(self):
        """
        Check if this expression is limited to 0 or 1 as value.

        @return: True if the value of this expression is either 0 or 1.
        """
        return False

    def type(self):
        """
        Determine the datatype of this expression.

        @return: A constant from the L{Type} class, representing the data type.
        """
        return Type.INTEGER


class ConstantNumeric(Expression):
    def __init__(self, value, pos=None):
        Expression.__init__(self, pos)
        self.value = generic.truncate_int32(value)
        self.uvalue = self.value
        if self.uvalue < 0:
            self.uvalue += 2**32

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Int:", self.value)

    def write(self, file, size):
        file.print_varx(self.value, size)

    def __str__(self):
        return str(self.value)

    def reduce(self, id_dicts=None, unknown_id_fatal=True):
        return self

    def supported_by_action2(self, raise_error):
        return True

    def supported_by_actionD(self, raise_error):
        return True

    def is_boolean(self):
        return self.value == 0 or self.value == 1

    def __eq__(self, other):
        return other is not None and isinstance(other, ConstantNumeric) and other.value == self.value

    def __hash__(self):
        return self.value


class ConstantFloat(Expression):
    def __init__(self, value, pos):
        Expression.__init__(self, pos)
        self.value = float(value)

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Float:", self.value)

    def __str__(self):
        return str(self.value)

    def reduce(self, id_dicts=None, unknown_id_fatal=True):
        return self

    def type(self):
        return Type.FLOAT



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

import hashlib
import os

from nml import generic, grfstrings, output_base


class OutputGRF(output_base.BinaryOutputBase):
    def __init__(self, filename):
        output_base.BinaryOutputBase.__init__(self, filename)
        self.encoder = None
        self.sprite_output = output_base.BinaryOutputBase(filename + ".sprite.tmp")
        self.md5 = hashlib.md5()
        # sprite_num is deliberately off-by-one because it is used as an
        # id between data and sprite section. For the sprite section an id
        # of 0 is invalid (means end of sprites), and for a non-NewGRF GRF
        # the first sprite is a real sprite.
        self.sprite_num = 1

    def open_file(self):
        # Remove / unlink the file, most useful for linux systems
        # See also issue #4165
        # If the file happens to be in use or non-existent, ignore
        try:
            os.unlink(self.filename)
        except OSError:
            # Ignore
            pass
        return open(self.filename, "wb")

    def get_md5(self):
        return self.md5.hexdigest()

    def assemble_file(self, real_file):
        # add end-of-chunks
        self.in_sprite = True
        self.print_dword(0)
        self.in_sprite = False
        self.sprite_output.in_sprite = True
        self.sprite_output.print_dword(0)
        self.sprite_output.in_sprite = False

        # add header
        header = bytearray([0x00, 0x00, ord("G"), ord("R"), ord("F"), 0x82, 0x0D, 0x0A, 0x1A, 0x0A])
        size = len(self.file) + 1
        header.append(size & 0xFF)
        header.append((size >> 8) & 0xFF)
        header.append((size >> 16) & 0xFF)
        header.append(size >> 24)
        header.append(0)  # no compression

        header_str = bytes(header)
        real_file.write(header_str)
        self.md5.update(header_str)

        # add data section, and then the sprite section
        real_file.write(self.file)
        self.md5.update(self.file)

        real_file.write(self.sprite_output.file)

    def open(self):
        output_base.BinaryOutputBase.open(self)
        self.sprite_output.open()

    def close(self):
        output_base.BinaryOutputBase.close(self)
        self.sprite_output.discard()

    def _print_utf8(self, char, stream):
        for c in chr(char).encode("utf8"):
            stream.print_byte(c)

    def print_string(self, value, final_zero=True, force_ascii=False, stream=None):
        if stream is None:
            stream = self

        if not grfstrings.is_ascii_string(value):
            if force_ascii:
                raise generic.ScriptError("Expected ascii string but got a unicode string")
            stream.print_byte(0xC3)
            stream.print_byte(0x9E)
        i = 0
        while i < len(value):
            if value[i] == "\\":
                if value[i + 1] in ("\\", '"'):
                    stream.print_byte(ord(value[i + 1]))
                    i += 2
                elif value[i + 1] == "U":
                    self._print_utf8(int(value[i + 2 : i + 6], 16), stream)
                    i += 6
                else:
                    stream.print_byte(int(value[i + 1 : i + 3], 16))
                    i += 3
            else:
                self._print_utf8(ord(value[i]), stream)
                i += 1
        if final_zero:
            stream.print_byte(0)

    def comment(self, msg):
        pass

    def start_sprite(self, size, is_real_sprite=False):
        if is_real_sprite:
            # Real sprite, this means no data is written to the data section
            # This call is still needed to open 'output mode'
            assert size == 0
            output_base.BinaryOutputBase.start_sprite(self, 9)
            self.print_dword(4)
            self.print_byte(0xFD)
            self.print_dword(self.sprite_num)
        else:
            output_base.BinaryOutputBase.start_sprite(self, size + 5)
            self.print_dword(size)
            self.print_byte(0xFF)

    def print_sprite(self, sprite_list):
        """
        @param sprite_list: List of non-empty real sprites for various bit depths / zoom levels
        @type  sprite_list: C{list} of L{RealSprite}
        """
        self.start_sprite(0, True)
        for sprite in sprite_list:
            self.print_single_sprite(sprite)
        self.end_sprite()

    def print_single_sprite(self, sprite_info):
        assert sprite_info.file is not None or sprite_info.mask_file is not None

        # Position for warning messages
        pos_warning = None
        if sprite_info.mask_file is not None:
            pos_warning = sprite_info.mask_file.pos
        elif sprite_info.file is not None:
            pos_warning = sprite_info.file.pos

        size_x, size_y, xoffset, yoffset, compressed_data, info_byte, crop_rect, warnings = self.encoder.get(
            sprite_info
        )

        for w in warnings:
            generic.print_warning(generic.Warning.GENERIC, w, pos_warning)

        self.sprite_output.start_sprite(len(compressed_data) + 18)
        self.wsprite_header(size_x, size_y, len(compressed_data), xoffset, yoffset, info_byte, sprite_info.zoom_level)
        self.sprite_output.print_data(compressed_data)
        self.sprite_output.end_sprite()

    def print_empty_realsprite(self):
        self.start_sprite(1)
        self.print_byte(0)
        self.end_sprite()

    def wsprite_header(self, size_x, size_y, size, xoffset, yoffset, info, zoom_level):
        self.sprite_output.print_dword(self.sprite_num)
        self.sprite_output.print_dword(size + 10)
        self.sprite_output.print_byte(info)
        self.sprite_output.print_byte(zoom_level)
        self.sprite_output.print_word(size_y)
        self.sprite_output.print_word(size_x)
        self.sprite_output.print_word(xoffset)
        self.sprite_output.print_word(yoffset)

    def print_named_filedata(self, filename):
        name = os.path.split(filename)[1]
        size = os.path.getsize(filename)

        self.start_sprite(0, True)
        self.sprite_output.start_sprite(8 + 3 + len(name) + 1 + size)

        self.sprite_output.print_dword(self.sprite_num)
        self.sprite_output.print_dword(3 + len(name) + 1 + size)
        self.sprite_output.print_byte(0xFF)
        self.sprite_output.print_byte(0xFF)
        self.sprite_output.print_byte(len(name))
        self.print_string(
            name, force_ascii=True, final_zero=True, stream=self.sprite_output
        )  # ASCII filenames seems sufficient.
        with open(generic.find_file(filename), "rb") as file:
            while True:
                data = file.read(1024)
                if len(data) == 0:
                    break
                for d in data:
                    self.sprite_output.print_byte(d)

        self.sprite_output.end_sprite()
        self.end_sprite()

    def end_sprite(self):
        output_base.BinaryOutputBase.end_sprite(self)
        self.sprite_num += 1



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic
from nml.actions import real_sprite
from nml.ast import base_statement, sprite_container


class BaseGraphics(base_statement.BaseStatement, sprite_container.SpriteContainer):
    """
    AST node for a 'base_graphics' block.
    NML syntax: base_graphics [block_name]([[sprite_num ,]default_file]) { ..real sprites.. }

    @ivar image_file: Default image file to use for sprites.
    @type image_file: C{None} if not specified, else L{StringLiteral}

    @ivar sprite_num: Sprite number of the first sprite (if provided explicitly)
    @type sprite_num: L{Expression} or C{None}

    @ivar sprite_list: List of real sprites to use
    @type sprite_list: Heterogeneous C{list} of L{RealSprite}, L{TemplateUsage}
    """

    def __init__(self, param_list, sprite_list, name, pos):
        base_statement.BaseStatement.__init__(self, "base_graphics-block", pos)
        sprite_container.SpriteContainer.__init__(self, "base_graphics-block", name)

        num_params = len(param_list)
        if not (0 <= num_params <= 2):
            raise generic.ScriptError(
                "base_graphics-block requires 0 to 2 parameters, encountered {:d}".format(num_params), pos
            )
        if num_params >= 2:
            self.sprite_num = param_list[0].reduce_constant()
        else:
            self.sprite_num = None

        if num_params >= 1:
            self.image_file = param_list[-1].reduce()
            if not isinstance(self.image_file, expression.StringLiteral):
                raise generic.ScriptError(
                    "The last base_graphics-block parameter 'file' must be a string literal", self.image_file.pos
                )
        else:
            self.image_file = None
        self.sprite_list = sprite_list
        self.add_sprite_data(self.sprite_list, self.image_file, pos)

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "base_graphics-block")
        generic.print_dbg(indentation + 2, "Source:", self.image_file.value if self.image_file is not None else "None")
        if self.block_name:
            generic.print_dbg(indentation + 2, "Name:", self.block_name)
        if self.sprite_num is not None:
            generic.print_dbg(indentation + 2, "Sprite number:", self.sprite_num)

        generic.print_dbg(indentation + 2, "Sprites:")
        for sprite in self.sprite_list:
            sprite.debug_print(indentation + 4)

    def get_action_list(self):
        actions = real_sprite.parse_sprite_data(self)
        actions[0].sprite_num = self.sprite_num
        return actions

    def __str__(self):
        name = str(self.block_name) if self.block_name is not None else ""
        params = [] if self.sprite_num is None else [self.sprite_num]
        if self.image_file is not None:
            params.append(self.image_file)
        ret = "base_graphics {}({}) {{\n".format(name, ", ".join(str(param) for param in params))
        for sprite in self.sprite_list:
            ret += "\t{}\n".format(sprite)
        ret += "}\n"
        return ret



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic
from nml.actions import action2, action2production, action2var
from nml.ast import base_statement

produce_base_class = action2.make_sprite_group_class(False, True, True)


class ProduceOld(produce_base_class):
    """
    AST node for a 'produce'-block, which is basically equivalent to the production callback.
    Syntax: produce(name, sub1, sub2, sub3, add1, add2[, again])
    @ivar param_list: List of parameters supplied to the produce-block.
                       - 0..2: Amounts of cargo to subtract from input
                       - 3..4: Amounts of cargo to add to output
                       - 5:    Run the production CB again if nonzero
    @type param_list: C{list} of L{Expression}
    """

    def __init__(self, param_list, pos):
        base_statement.BaseStatement.__init__(self, "produce-block", pos, False, False)
        if not (6 <= len(param_list) <= 7):
            raise generic.ScriptError(
                "produce-block requires 6 or 7 parameters, encountered {:d}".format(len(param_list)), self.pos
            )
        name = param_list[0]
        if not isinstance(name, expression.Identifier):
            raise generic.ScriptError("produce parameter 1 'name' should be an identifier.", name.pos)
        self.initialize(name, 0x0A)
        self.param_list = param_list[1:]
        if len(self.param_list) < 6:
            self.param_list.append(expression.ConstantNumeric(0))

    def pre_process(self):
        generic.print_warning(
            generic.Warning.DEPRECATION,
            "Consider using the new produce() syntax for '{}'".format(self.name),
            self.name.pos,
        )
        var_scope = action2var.get_scope(0x0A)
        for i, param in enumerate(self.param_list):
            self.param_list[i] = action2var.reduce_varaction2_expr(param, var_scope)
        produce_base_class.pre_process(self)

    def collect_references(self):
        all_refs = []
        for expr in self.param_list:
            all_refs += expr.collect_references()
        return all_refs

    def __str__(self):
        return "produce({});\n".format(", ".join(str(x) for x in [self.name] + self.param_list))

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Produce, name =", self.name)
        generic.print_dbg(indentation + 2, "Subtract from input:")
        for expr in self.param_list[0:3]:
            expr.debug_print(indentation + 4)
        generic.print_dbg(indentation + 2, "Add to output:")
        for expr in self.param_list[3:5]:
            expr.debug_print(indentation + 4)
        generic.print_dbg(indentation + 2, "Again:")
        self.param_list[5].debug_print(indentation + 4)

    def get_action_list(self):
        if self.prepare_act2_output():
            return action2production.get_production_actions(self)
        return []


class Produce(produce_base_class):
    """
    AST node for a 'produce'-block for version 2 industry production callback.
    Syntax: produce(name, [INP1: expression; ...], [OUT1: expression; ...], again)
        or: produce(name, [INP1: expression; ...], [OUT1: expression; ...])

    @ivar name: name given to this produce block
    @type name: L{Identifier}

    @ivar subtract_in: List of input cargos that are consumed for this production
                       (cargo label to register-expression mapping)
    @type subtract_in: C{list} of L{Assignment}

    @ivar add_out: List of output cargos that are generated for this production
                   (cargo label to register-expression mapping)
    @type add_out: C{list} of L{Assignment}

    @ivar again: Production CB is repeated if this is a register evaluating to nonzero
    @type again: L{Expression}
    """

    def __init__(self, name, subtract_in, add_out, again, pos):
        if not isinstance(name, expression.Identifier):
            raise generic.ScriptError("produce parameter 1 'name' should be an identifier.", name.pos)
        base_statement.BaseStatement.__init__(self, "produce-block", pos, False, False)
        self.initialize(name, 0x0A)
        self.subtract_in = subtract_in
        self.add_out = add_out
        self.again = again

    def pre_process(self):
        var_scope = action2var.get_scope(0x0A)
        for i, param in enumerate(self.subtract_in):
            self.subtract_in[i].value = action2var.reduce_varaction2_expr(param.value, var_scope)
        for i, param in enumerate(self.add_out):
            self.add_out[i].value = action2var.reduce_varaction2_expr(param.value, var_scope)
        self.again = action2var.reduce_varaction2_expr(self.again, var_scope)
        produce_base_class.pre_process(self)

    def collect_references(self):
        all_refs = []
        for cargopair in self.subtract_in + self.add_out:
            all_refs += cargopair.value.collect_references()
        return all_refs

    def __str__(self):
        return "produce({0}, [{1}], [{2}], {3})\n".format(
            str(self.name),
            " ".join(str(x) for x in self.subtract_in),
            " ".join(str(x) for x in self.add_out),
            str(self.again),
        )

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Produce, name =", self.name)
        for cargopair in self.subtract_in:
            generic.print_dbg(indentation + 2, "Subtract from input {0}:".format(cargopair.name.value))
            cargopair.value.debug_print(indentation + 4)
        for cargopair in self.add_out:
            generic.print_dbg(indentation + 2, "Add to output {0}:".format(cargopair.name.value))
            cargopair.value.debug_print(indentation + 4)
        generic.print_dbg(indentation + 2, "Again:")
        self.again.debug_print(indentation + 4)

    def get_action_list(self):
        if self.prepare_act2_output():
            return action2production.get_production_v2_actions(self)
        return []



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

import array

from nml import generic, lz77, palette, spritecache
from nml.actions import real_sprite

try:
    from PIL import Image
except ImportError:
    # Image is required only when using graphics
    pass

# Some constants for the 'info' byte
INFO_RGB = 1
INFO_ALPHA = 2
INFO_PAL = 4
INFO_TILE = 8
INFO_NOCROP = 0x40


def get_bpp(info):
    bpp = 0
    if (info & INFO_RGB) != 0:
        bpp += 3
    if (info & INFO_ALPHA) != 0:
        bpp += 1
    if (info & INFO_PAL) != 0:
        bpp += 1
    return bpp


def has_transparency(info):
    return (info & (INFO_ALPHA | INFO_PAL)) != 0


def transparency_offset(info):
    """
    Determine which byte within a pixel should be 0 for a pixel to be transparent
    """
    assert has_transparency(info)
    # There is an alpha or palette component, or both, else no transparency at all
    # Pixel is transparent if either:
    # - alpha channel present and 0
    # - palette present and 0, and alpha not present
    # In either case, the pixel is transparent,
    # if byte 3 (with RGB present) or byte 0 (without RGB) is 0
    if (info & INFO_RGB) != 0:
        return 3
    else:
        return 0


class SpriteEncoder:
    """
    Algorithms for cropping and compressing sprites. That is encoding source images into GRF sprites.

    @ivar compress_grf: Compress sprites.
    @type compress_grf: C{bool}

    @ivar crop_sprites: Crop sprites if possible.
    @type crop_sprites: C{bool}

    @ivar palette: Palette for encoding, see L{palette.palette_name}.
    @type palette: C{str}

    @ivar cached_image_files: Currently opened source image files.
    @type cached_image_files: C{dict} mapping C{str} to C{Image}
    """

    def __init__(self, compress_grf, crop_sprites, palette):
        self.compress_grf = compress_grf
        self.crop_sprites = crop_sprites
        self.palette = palette
        self.sprite_cache = spritecache.SpriteCache()
        self.cached_image_files = {}

    def open(self, sprite_files):
        """
        Start the encoder, read caches, and stuff.

        @param sprite_files: List of sprites per source image file.
        @type  sprite_files: C{dict} that maps (C{tuple} of C{str}) to (C{RealSprite})
        """
        num_sprites = sum(len(sprite_list) for sprite_list in sprite_files.values())

        generic.print_progress("Encoding ...")

        num_cached = 0
        num_dup = 0
        num_enc = 0
        num_orphaned = 0
        count_sprites = 0
        for sources, sprite_list in sprite_files.items():
            # Iterate over sprites grouped by source image file.
            #  - Open source files only once. (speed)
            #  - Do not keep files around for long. (memory)

            source_name = "_".join(src for src in sources if src is not None)

            local_cache = spritecache.SpriteCache(sources)
            local_cache.read_cache()

            for sprite_info in sprite_list:
                count_sprites += 1
                generic.print_progress(
                    "Encoding {}/{}: {}".format(count_sprites, num_sprites, source_name), incremental=True
                )

                cache_key = sprite_info.get_cache_key(self.crop_sprites)
                cache_item = local_cache.get_item(cache_key, self.palette)

                in_use = False
                in_old_cache = False
                if cache_item is not None:
                    # Write a sprite from the cached data
                    compressed_data, info_byte, crop_rect, pixel_stats, in_old_cache, in_use = cache_item
                    if in_use:
                        num_dup += 1
                    else:
                        num_cached += 1
                else:
                    (
                        size_x,
                        size_y,
                        xoffset,
                        yoffset,
                        compressed_data,
                        info_byte,
                        crop_rect,
                        pixel_stats,
                    ) = self.encode_sprite(sprite_info)
                    num_enc += 1

                # Store sprite in cache, unless already up-to-date
                if not in_use:
                    cache_item = (compressed_data, info_byte, crop_rect, pixel_stats, in_old_cache, True)
                    local_cache.add_item(cache_key, self.palette, cache_item)

            # Delete all files from dictionary to free memory
            self.cached_image_files.clear()

            num_orphaned += local_cache.count_orphaned()

            # Only write cache if compression is enabled. Uncompressed data is not worth to be cached.
            if self.compress_grf:
                local_cache.write_cache()

            # Transfer data to global cache for later usage
            self.sprite_cache.cached_sprites.update(local_cache.cached_sprites)

        generic.print_progress("Encoding ...", incremental=True)
        generic.clear_progress()
        generic.print_info(
            "{} sprites, {} cached, {} orphaned, {} duplicates, {} newly encoded ({})".format(
                num_sprites, num_cached, num_orphaned, num_dup, num_enc, "native" if lz77.is_native else "python"
            )
        )

    def close(self):
        """
        Close the encoder, validate data, write caches, and stuff.
        """
        pass

    def get(self, sprite_info):
        """
        Get encoded sprite date, either from cache, or new.

        @param sprite_info: Sprite metadata
        @type  sprite_info: C{RealSprite}
        """

        # Try finding the file in the cache
        # open() should have put all sprites into it.
        cache_key = sprite_info.get_cache_key(self.crop_sprites)
        cache_item = self.sprite_cache.get_item(cache_key, self.palette)

        assert cache_item is not None

        compressed_data, info_byte, crop_rect, pixel_stats, in_old_cache, in_use = cache_item

        size_x = sprite_info.xsize.value
        size_y = sprite_info.ysize.value
        xoffset = sprite_info.xrel.value
        yoffset = sprite_info.yrel.value
        if cache_key[-1]:
            size_x, size_y, xoffset, yoffset = self.recompute_offsets(size_x, size_y, xoffset, yoffset, crop_rect)

        warnings = []
        total = pixel_stats.get("total", 0)
        if total > 0:
            if cache_key[0] is not None:
                image_32_pos = generic.PixelPosition(cache_key[0], cache_key[1][0], cache_key[1][1])
                alpha = pixel_stats.get("alpha", 0)
                if alpha > 0 and (sprite_info.flags.value & real_sprite.FLAG_NOALPHA) != 0:
                    warnings.append(
                        "{}: {:d} of {:d} pixels ({:d}%) are semi-transparent, but NOALPHA is in flags".format(
                            str(image_32_pos), alpha, total, alpha * 100 // total
                        )
                    )

            if cache_key[2] is not None:
                image_8_pos = generic.PixelPosition(cache_key[2], cache_key[3][0], cache_key[3][1])
                white = pixel_stats.get("white", 0)
                anim = pixel_stats.get("anim", 0)
                if white > 0 and (sprite_info.flags.value & real_sprite.FLAG_WHITE) == 0:
                    warnings.append(
                        "{}: {:d} of {:d} pixels ({:d}%) are pure white, but WHITE isn't in flags".format(
                            str(image_8_pos), white, total, white * 100 // total
                        )
                    )
                if anim > 0 and (sprite_info.flags.value & real_sprite.FLAG_ANIM) == 0:
                    warnings.append(
                        "{}: {:d} of {:d} pixels ({:d}%) are animated, but ANIM isn't in flags".format(
                            str(image_8_pos), anim, total, anim * 100 // total
                        )
                    )

        return (size_x, size_y, xoffset, yoffset, compressed_data, info_byte, crop_rect, warnings)

    def open_image_file(self, filename):
        """
        Obtain a handle to an image file

        @param filename: Name of the file
        @type  filename: C{str}

        @return: Image file
        @rtype:  L{Image}
        """
        if filename in self.cached_image_files:
            im = self.cached_image_files[filename]
        else:
            im = Image.open(generic.find_file(filename))
            self.cached_image_files[filename] = im
        return im

    def encode_sprite(self, sprite_info):
        """
        Crop and compress a real sprite.

        @param sprite_info: Sprite metadata
        @type  sprite_info: C{RealSprite}

        @return: size_x, size_y, xoffset, yoffset, compressed_data, info_byte, crop_rect, pixel_stats
        @rtype: C{tuple}
        """

        filename_8bpp = None
        filename_32bpp = None
        if sprite_info.bit_depth == 8:
            filename_8bpp = sprite_info.file
        else:
            filename_32bpp = sprite_info.file
            filename_8bpp = sprite_info.mask_file

        # Get initial info_byte and dimensions from sprite_info.
        # These values will be changed depending on cropping and compression.
        info_byte = INFO_NOCROP if (sprite_info.flags.value & real_sprite.FLAG_NOCROP) != 0 else 0
        size_x = sprite_info.xsize.value
        size_y = sprite_info.ysize.value
        xoffset = sprite_info.xrel.value
        yoffset = sprite_info.yrel.value

        # Select region of image bounded by x/ypos and x/ysize
        x = sprite_info.xpos.value
        y = sprite_info.ypos.value
        if sprite_info.bit_depth == 8 or sprite_info.mask_pos is None:
            mask_x, mask_y = x, y
        else:
            mask_x = sprite_info.mask_pos[0].value
            mask_y = sprite_info.mask_pos[1].value

        pixel_stats = {"total": size_x * size_y, "alpha": 0, "white": 0, "anim": 0}

        # Read and validate image data
        if filename_32bpp is not None:
            im = self.open_image_file(filename_32bpp.value)
            if im.mode not in ("RGB", "RGBA"):
                pos = generic.build_position(sprite_info.poslist)
                raise generic.ImageError("32bpp image is not a full colour RGB(A) image.", filename_32bpp.value, pos)
            info_byte |= INFO_RGB
            if im.mode == "RGBA":
                info_byte |= INFO_ALPHA

            (im_width, im_height) = im.size
            if x < 0 or y < 0 or x + size_x > im_width or y + size_y > im_height:
                pos = generic.build_position(sprite_info.poslist)
                raise generic.ScriptError("Read beyond bounds of image file '{}'".format(filename_32bpp.value), pos)
            try:
                sprite = im.crop((x, y, x + size_x, y + size_y))
            except OSError:
                pos = generic.build_position(sprite_info.poslist)
                raise generic.ImageError("Failed to crop 32bpp {} image".format(im.format), filename_32bpp.value, pos)
            rgb_sprite_data = sprite.tobytes()

            if (info_byte & INFO_ALPHA) != 0:
                # Check for half-transparent pixels (not valid for ground sprites)
                pixel_stats["alpha"] = sum(0x00 < p < 0xFF for p in rgb_sprite_data[3::4])

        if filename_8bpp is not None:
            mask_im = self.open_image_file(filename_8bpp.value)
            if mask_im.mode != "P":
                pos = generic.build_position(sprite_info.poslist)
                raise generic.ImageError("8bpp image does not have a palette", filename_8bpp.value, pos)
            im_mask_pal = palette.validate_palette(mask_im, filename_8bpp.value)
            info_byte |= INFO_PAL

            (im_width, im_height) = mask_im.size
            if mask_x < 0 or mask_y < 0 or mask_x + size_x > im_width or mask_y + size_y > im_height:
                pos = generic.build_position(sprite_info.poslist)
                raise generic.ScriptError("Read beyond bounds of image file '{}'".format(filename_8bpp.value), pos)
            try:
                mask_sprite = mask_im.crop((mask_x, mask_y, mask_x + size_x, mask_y + size_y))
            except OSError:
                pos = generic.build_position(sprite_info.poslist)
                raise generic.ImageError(
                    "Failed to crop 8bpp {} image".format(mask_im.format), filename_8bpp.value, pos
                )
            mask_sprite_data = self.palconvert(mask_sprite.tobytes(), im_mask_pal)

            # Check for white pixels; those that cause "artefacts" when shading
            pixel_stats["white"] = sum(p == 255 for p in mask_sprite_data)

            # Check for palette animation colours
            if self.palette == "DEFAULT":
                pixel_stats["anim"] = sum(0xE3 <= p <= 0xFE for p in mask_sprite_data)
            else:
                pixel_stats["anim"] = sum(0xD9 <= p <= 0xF4 for p in mask_sprite_data)

        # Compose pixel information in an array of bytes
        sprite_data = array.array("B")
        if (info_byte & INFO_RGB) != 0 and (info_byte & INFO_PAL) != 0:
            mask_data = array.array("B", mask_sprite_data)  # Convert to numeric
            rgb_data = array.array("B", rgb_sprite_data)
            if (info_byte & INFO_ALPHA) != 0:
                for i in range(len(mask_sprite_data)):
                    sprite_data.extend(rgb_data[4 * i : 4 * (i + 1)])
                    sprite_data.append(mask_data[i])
            else:
                for i in range(len(mask_sprite_data)):
                    sprite_data.extend(rgb_data[3 * i : 3 * (i + 1)])
                    sprite_data.append(mask_data[i])
        elif (info_byte & INFO_RGB) != 0:
            sprite_data.frombytes(rgb_sprite_data)
        else:
            sprite_data.frombytes(mask_sprite_data)

        bpp = get_bpp(info_byte)
        assert len(sprite_data) == size_x * size_y * bpp
        if self.crop_sprites and ((info_byte & INFO_NOCROP) == 0):
            sprite_data, crop_rect = self.crop_sprite(sprite_data, size_x, size_y, info_byte, bpp)
            size_x, size_y, xoffset, yoffset = self.recompute_offsets(size_x, size_y, xoffset, yoffset, crop_rect)
        else:
            crop_rect = None
        assert len(sprite_data) == size_x * size_y * bpp

        compressed_data = self.sprite_compress(sprite_data)
        # Try tile compression, and see if it results in a smaller file size
        tile_data = self.sprite_encode_tile(size_x, size_y, sprite_data, info_byte, bpp)
        if tile_data is not None:
            tile_compressed_data = self.sprite_compress(tile_data)
            # Tile compression adds another 4 bytes for the uncompressed chunked data in the header
            if len(tile_compressed_data) + 4 < len(compressed_data):
                info_byte |= INFO_TILE
                data_len = len(tile_data)
                compressed_data = array.array("B")
                compressed_data.append(data_len & 0xFF)
                compressed_data.append((data_len >> 8) & 0xFF)
                compressed_data.append((data_len >> 16) & 0xFF)
                compressed_data.append((data_len >> 24) & 0xFF)
                compressed_data.extend(tile_compressed_data)

        return (size_x, size_y, xoffset, yoffset, compressed_data, info_byte, crop_rect, pixel_stats)

    def fakecompress(self, data):
        i = 0
        output = array.array("B")
        length = len(data)
        while i < length:
            n = min(length - i, 127)
            output.append(n)
            output.extend(data[i : i + n])
            i += n
        return output

    def sprite_compress(self, data):
        if self.compress_grf:
            stream = lz77.encode(data)
        else:
            stream = self.fakecompress(data)
        return stream

    def sprite_encode_tile(self, size_x, size_y, data, info, bpp, long_format=False):
        long_chunk = size_x > 256

        # There are basically four different encoding configurations here,
        # but just two variables. If the width of the sprite is more than
        # 256, then the chunk could be 'out of bounds' and thus the long
        # chunk format is used. If the sprite is more than 65536 bytes,
        # then the offsets might not fit and the long format method is
        # used. The latter is enabled via recursion if it's needed.
        if not has_transparency(info):
            return None
        trans_offset = transparency_offset(info)
        max_chunk_len = 0x7FFF if long_chunk else 0x7F
        line_offset_size = 4 if long_format else 2  # Whether to use 2 or 4 bytes in the list of line offsets
        output = array.array("B", [0] * (line_offset_size * size_y))

        for y in range(size_y):
            # Write offset in the correct place, in little-endian format
            offset = len(output)
            output[y * line_offset_size] = offset & 0xFF
            output[y * line_offset_size + 1] = (offset >> 8) & 0xFF
            if long_format:
                output[y * line_offset_size + 2] = (offset >> 16) & 0xFF
                output[y * line_offset_size + 3] = (offset >> 24) & 0xFF

            line_start = y * size_x * bpp

            line_parts = []
            x1 = 0
            while True:
                # Skip transparent pixels
                while x1 < size_x and data[line_start + x1 * bpp + trans_offset] == 0:
                    x1 += 1
                if x1 == size_x:
                    # End-of-line reached
                    break

                # Grab as many non-transparent pixels as possible, but not without x2-x1 going out of bounds
                # Only stop the chunk when encountering 3 consecutive transparent pixels
                x2 = x1 + 1
                while x2 - x1 < max_chunk_len and (
                    (x2 < size_x and data[line_start + x2 * bpp + trans_offset] != 0)
                    or (x2 + 1 < size_x and data[line_start + (x2 + 1) * bpp + trans_offset] != 0)
                    or (x2 + 2 < size_x and data[line_start + (x2 + 2) * bpp + trans_offset] != 0)
                ):
                    x2 += 1
                line_parts.append((x1, x2))
                x1 = x2

            if len(line_parts) == 0:
                # Completely transparent line
                if long_chunk:
                    output.extend((0, 0x80, 0, 0))
                else:
                    output.extend((0x80, 0))
                continue

            for idx, part in enumerate(line_parts):
                x1, x2 = part
                last_mask = 0x80 if idx == len(line_parts) - 1 else 0
                chunk_len = x2 - x1
                if long_chunk:
                    output.extend((chunk_len & 0xFF, (chunk_len >> 8) | last_mask, x1 & 0xFF, x1 >> 8))
                else:
                    output.extend((chunk_len | last_mask, x1))
                output.extend(data[line_start + x1 * bpp : line_start + x2 * bpp])

        if len(output) > 65535 and not long_format:
            # Recurse into the long format if that's possible.
            return self.sprite_encode_tile(size_x, size_y, data, info, bpp, True)
        return output

    def recompute_offsets(self, size_x, size_y, xoffset, yoffset, crop_rect):
        # Recompute sizes and offsets after cropping a sprite
        left, right, top, bottom = crop_rect
        size_x -= left + right
        size_y -= top + bottom
        xoffset += left
        yoffset += top
        return size_x, size_y, xoffset, yoffset

    def crop_sprite(self, data, size_x, size_y, info, bpp):
        left, right, top, bottom = 0, 0, 0, 0
        if not has_transparency(info):
            return (data, (left, right, top, bottom))

        trans_offset = transparency_offset(info)
        line_size = size_x * bpp  # size (no. of bytes) of a scan line
        data_size = len(data)

        # Crop the top of the sprite
        while size_y > 1 and not any(data[line_size * top + trans_offset : line_size * (top + 1) : bpp]):
            top += 1
            size_y -= 1

        # Crop the bottom of the sprite
        while size_y > 1 and not any(
            data[data_size - line_size * (bottom + 1) + trans_offset : data_size - line_size * bottom : bpp]
        ):
            # Don't use negative indexing, it breaks for the last line (where you'd need index 0)
            bottom += 1
            size_y -= 1

        # Modify data by removing top/bottom
        data = data[line_size * top : data_size - line_size * bottom]

        # Crop the left of the sprite
        while size_x > 1 and not any(data[left * bpp + trans_offset :: line_size]):
            left += 1
            size_x -= 1

        # Crop the right of the sprite
        while size_x > 1 and not any(data[line_size - (right + 1) * bpp + trans_offset :: line_size]):
            right += 1
            size_x -= 1

        # Removing left/right data is not easily done by slicing
        # Best to create a new array
        if left + right > 0:
            new_data = array.array("B")
            for y in range(0, size_y):
                a = data[y * line_size + left * bpp : (y + 1) * line_size - right * bpp]
                new_data.extend(a)
            data = new_data

        return (data, (left, right, top, bottom))

    def palconvert(self, sprite_str, orig_pal):
        if orig_pal == "LEGACY" and self.palette == "DEFAULT":
            return sprite_str.translate(real_sprite.translate_w2d)
        else:
            return sprite_str



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

import os
import sys


def get_lib_versions():
    versions = {}
    # PIL
    try:
        import PIL

        versions["PIL"] = PIL.__version__
    except ImportError:
        versions["PIL"] = "Not found!"

    # PLY
    try:
        from ply import lex

        versions["PLY"] = lex.__version__
    except ImportError:
        versions["PLY"] = "Not found!"

    return versions


def get_nml_version():
    # First check if this is a git repository, and use that version if available.
    # (unless this is a released tarball, see below)
    try:
        from nml import version_update

        version = version_update.get_git_version()
        if version:
            return version
    except ImportError:
        # version_update is excluded from release tarballs,
        #  so that the predetermined version is always used.
        pass

    # No repository was found. Return the version which was saved upon build.
    try:
        from nml import __version__

        version = __version__.version
    except ImportError:
        version = "unknown"
    return version


def get_cli_version():
    # Version string for usage in command line
    result = get_nml_version() + "\n\n"

    nmlc_path = os.path.abspath(sys.argv[0])
    result += "nmlc: {}\n".format(nmlc_path)

    from nml import lz77

    lz77_ver = "C (native)" if lz77.is_native else "Python"
    result += "LZ77 implementation: {}\n\n".format(lz77_ver)

    result += "Library versions encountered:\n"
    for lib, lib_ver in get_lib_versions().items():
        result += "  {}: {}\n".format(lib, lib_ver)

    result += "\nPython: {}\n".format(sys.executable)
    result += "version {}".format(sys.version)
    return result



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml.actions import base_action


class ActionE(base_action.BaseAction):
    def __init__(self, grfid_list):
        self.grfid_list = grfid_list

    def write(self, file):
        size = 2 + 4 * len(self.grfid_list)
        file.start_sprite(size)
        file.print_bytex(0x0E)
        file.print_byte(len(self.grfid_list))
        for grfid in self.grfid_list:
            file.newline()
            file.print_dwordx(grfid)
        file.newline()
        file.end_sprite()


def parse_deactivate_block(block):
    return [ActionE(block.grfid_list)]



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic, global_constants, nmlop
from nml.actions import action2, action2real, action2var_variables, action4, action6, actionD
from nml.ast import general, switch, spriteblock


class Action2Var(action2.Action2):
    """
    Variational Action2. This is the NFO equivalent of a switch-block in NML.
    It computes a single integer from one or more variables and picks it's
    return value based on the result of the computation. The return value can
    be either a 15bit integer or a reference to another action2.

    @ivar type_byte: The size (byte, word, double word) and access type (own
                     object or related object). 0x89 (own object, double word)
                     and 0x8A (related object, double word) and the only
                     supported values.
    @type type_byte: C{int}

    @ivar ranges: List of return value ranges. Each range contains a minimum and
                  a maximum value and a return value. The list is checked in order,
                  if the result of the computation is between the minimum and
                  maximum (inclusive) of one range the result of that range is
                  returned. The result can be either an integer of another
                  action2.
    @ivar ranges: C{list} of L{VarAction2Range}
    """

    def __init__(self, feature, name, pos, type_byte, param_registers=None):
        action2.Action2.__init__(self, feature, name, pos)
        self.type_byte = type_byte
        self.ranges = []
        self.param_registers = param_registers or []

    def resolve_tmp_storage(self):
        # A return action may use the parameters of its parent
        # Make sure param registers are not reused
        for var in self.var_list:
            if isinstance(var, VarAction2LoadCallParam):
                self.remove_tmp_location(var.parameter, False)

        for var in self.param_registers + self.var_list:  # Allocate param registers first
            if isinstance(var, (VarAction2StoreTempVar, VarAction2CallParam)):
                if isinstance(var, VarAction2CallParam) and var.register:
                    continue
                if not self.tmp_locations:
                    raise generic.ScriptError(
                        "There are not enough registers available "
                        + "to perform all required computations in switch blocks. "
                        + "Please reduce the complexity of your code.",
                        self.pos,
                    )
                location = self.tmp_locations[0]
                self.remove_tmp_location(location, False)
                var.set_register(location)

    def prepare_output(self, sprite_num):
        action2.Action2.prepare_output(self, sprite_num)
        for i in range(0, len(self.var_list) - 1, 2):
            self.var_list[i].shift |= 0x20

        for i in range(0, len(self.var_list), 2):
            if isinstance(self.var_list[i], VarAction2ProcCallVar):
                self.var_list[i].resolve_parameter(self.feature)

        for r in self.ranges:
            if isinstance(r.result, expression.SpriteGroupRef):
                r.result = r.result.get_action2_id(self.feature)
            else:
                r.result = r.result.value | 0x8000
        if isinstance(self.default_result, expression.SpriteGroupRef):
            self.default_result = self.default_result.get_action2_id(self.feature)
        else:
            self.default_result = self.default_result.value | 0x8000

    def write(self, file):
        # type_byte, num_ranges, default_result = 4
        # 2 bytes for the result, 8 bytes for the min/max range.
        size = 4 + (2 + 8) * len(self.ranges)
        for var in self.var_list:
            if isinstance(var, nmlop.Operator):
                size += 1
            else:
                size += var.get_size()

        regs = ["{} : register {:X}".format(reg.name, reg.register) for reg in self.param_registers]
        self.write_sprite_start(file, size, regs)
        file.print_bytex(self.type_byte)
        file.newline()
        for var in self.var_list:
            if isinstance(var, nmlop.Operator):
                file.print_bytex(var.act2_num, var.act2_str)
            else:
                var.write(file, 4)
                file.newline(var.comment)
        file.print_byte(len(self.ranges))
        file.newline()
        for r in self.ranges:
            file.print_wordx(r.result)
            file.print_varx(r.min.value, 4)
            file.print_varx(r.max.value, 4)
            file.newline(r.comment)
        file.print_wordx(self.default_result)
        file.comment(self.default_comment)
        file.end_sprite()


class VarAction2Var:
    """
    Represents a variable for use in a (advanced) variational action2.

    @ivar var_num: Number of the variable to use.
    @type var_num: C{int}

    @ivar shift: The number of bits to shift the value of the given variable to the right.
    @type shift: C{int}

    @ivar mask: Bitmask to use on the value after shifting it.
    @type mask: C{int}

    @ivar parameter: Parameter to be used as argument for the variable.
    @type parameter: C{int} or C{None}
    @precondition: (0x60 <= var_num <= 0x7F) == (parameter is not None)

    @ivar add: If not C{None}, add this value to the result.
    @type add: C{int} or C{None}

    @ivar div: If not C{None}, divide the result by this.
    @type add: C{int} or C{None}

    @ivar mod: If not C{None}, compute (result module mod).
    @type add: C{int} or C{None}

    @ivar comment: Textual description of this variable.
    @type comment: C{basestr}
    """

    def __init__(self, var_num, shift, mask, parameter=None, comment=""):
        self.var_num = var_num
        self.shift = shift
        self.mask = mask
        self.parameter = parameter
        self.add = None
        self.div = None
        self.mod = None
        self.comment = comment

    def write(self, file, size):
        file.print_bytex(self.var_num)
        if self.parameter is not None:
            file.print_bytex(self.parameter)
        if self.mod is not None:
            self.shift |= 0x80
        elif self.add is not None or self.div is not None:
            self.shift |= 0x40
        file.print_bytex(self.shift)
        file.print_varx(self.mask, size)
        if self.add is not None:
            file.print_varx(self.add, size)
            if self.div is not None:
                file.print_varx(self.div, size)
            elif self.mod is not None:
                file.print_varx(self.mod, size)
            else:
                # no div or add, just divide by 1
                file.print_varx(1, size)

    def get_size(self):
        # var number (1) [+ parameter (1)] + shift num (1) + and mask (4) [+ add (4) + div/mod (4)]
        size = 6
        if self.parameter is not None:
            size += 1
        if self.add is not None or self.div is not None or self.mod is not None:
            size += 8
        return size

    def supported_by_actionD(self, raise_error):
        assert not raise_error
        return False


# Class for var 7E procedure calls
class VarAction2ProcCallVar(VarAction2Var):
    def __init__(self, sg_ref):
        if not sg_ref.act2:
            if not isinstance(action2.resolve_spritegroup(sg_ref.name), (switch.Switch, switch.RandomSwitch)):
                raise generic.ScriptError(
                    "Block with name '{}' is not a valid procedure".format(sg_ref.name), sg_ref.pos
                )
            if not sg_ref.is_procedure:
                raise generic.ScriptError("Unexpected identifier encountered: '{}'".format(sg_ref.name), sg_ref.pos)
        VarAction2Var.__init__(self, 0x7E, 0, 0, comment=str(sg_ref))
        # Reference to the called action2
        self.sg_ref = sg_ref

    def resolve_parameter(self, feature):
        self.parameter = self.sg_ref.get_action2_id(feature)

    def get_size(self):
        return 7

    def write(self, file, size):
        self.mask = get_mask(size)
        VarAction2Var.write(self, file, size)


# General load and store class for temp parameters
# Register is allocated at the store operation
class VarAction2StoreTempVar(VarAction2Var):
    def __init__(self):
        VarAction2Var.__init__(self, 0x1A, 0, 0)
        # mask holds the number, it's resolved in Action2Var.resolve_tmp_storage
        self.load_vars = []

    def set_register(self, register):
        self.mask = register
        for load_var in self.load_vars:
            load_var.parameter = register

    def get_size(self):
        return 6


def get_mask(size):
    if size == 1:
        return 0xFF
    elif size == 2:
        return 0xFFFF
    return 0xFFFFFFFF


class VarAction2LoadTempVar(VarAction2Var, expression.Expression):
    def __init__(self, tmp_var):
        VarAction2Var.__init__(self, 0x7D, 0, 0)
        expression.Expression.__init__(self, None)
        assert isinstance(tmp_var, VarAction2StoreTempVar)
        tmp_var.load_vars.append(self)

    def write(self, file, size):
        self.mask = get_mask(size)
        VarAction2Var.write(self, file, size)

    def get_size(self):
        return 7

    def reduce(self, id_dicts=None, unknown_id_fatal=True):
        return self

    def supported_by_action2(self, raise_error):
        return True

    def supported_by_actionD(self, raise_error):
        assert not raise_error
        return False


# Temporary load and store classes used for spritelayout parameters
# Register is allocated in a separate entity
class VarAction2CallParam:
    def __init__(self, name):
        self.register = None
        self.store_vars = []
        self.load_vars = []
        self.name = name

    def set_register(self, register):
        self.register = register
        for store_var in self.store_vars:
            store_var.mask = register
        for load_var in self.load_vars:
            load_var.parameter = register


class VarAction2LoadCallParam(VarAction2Var, expression.Expression):
    def __init__(self, param, name):
        assert isinstance(param, VarAction2CallParam)
        VarAction2Var.__init__(self, 0x7D, 0, 0, comment=param.name)
        expression.Expression.__init__(self, None)
        assert isinstance(param, VarAction2CallParam)
        param.load_vars.append(self)
        self.name = name
        # Register is stored in parameter

    def write(self, file, size):
        self.mask = get_mask(size)
        VarAction2Var.write(self, file, size)

    def get_size(self):
        return 7

    def reduce(self, id_dicts=None, unknown_id_fatal=True):
        return self

    def supported_by_action2(self, raise_error):
        return True

    def supported_by_actionD(self, raise_error):
        assert not raise_error
        return False

    def __str__(self):
        return self.name


class VarAction2StoreCallParam(VarAction2Var):
    def __init__(self, param):
        VarAction2Var.__init__(self, 0x1A, 0, 0)
        assert isinstance(param, VarAction2CallParam)
        param.store_vars.append(self)
        # Register is stored in mask

    def get_size(self):
        return 6


class VarAction2Range:
    def __init__(self, min, max, result, comment):
        self.min = min
        self.max = max
        self.result = result
        self.comment = comment


class Modification:
    def __init__(self, param, size, offset):
        self.param = param
        self.size = size
        self.offset = offset


class Varaction2Parser:
    def __init__(self, action_feature, var_range=0x89):
        self.action_feature = action_feature
        self.var_range = var_range
        self.var_scope = get_scope(action_feature, var_range)  # Depends on action_feature and var_range
        self.extra_actions = []
        self.mods = []
        self.var_list = []
        self.var_list_size = 0
        self.proc_call_list = []
        self.in_store_op = False

    def preprocess_binop(self, expr):
        """
        Several nml operators are not directly support by nfo so we have to work
        around that by implementing those operators in terms of others.

        @return: A pre-processed version of the expression.
        @rtype:  L{Expression}
        """
        assert isinstance(expr, expression.BinOp)
        if expr.op == nmlop.CMP_LT:
            # return value is 0, 1 or 2, we want to map 0 to 1 and the others to 0
            expr = nmlop.VACT2_CMP(expr.expr1, expr.expr2)
            # reduce the problem to 0/1
            expr = nmlop.MIN(expr, 1)
            # and invert the result
            expr = nmlop.XOR(expr, 1)
        elif expr.op == nmlop.CMP_GT:
            # return value is 0, 1 or 2, we want to map 2 to 1 and the others to 0
            expr = nmlop.VACT2_CMP(expr.expr1, expr.expr2)
            # subtract one
            expr = nmlop.SUB(expr, 1)
            # map -1 and 0 to 0
            expr = nmlop.MAX(expr, 0)
        elif expr.op == nmlop.CMP_LE:
            # return value is 0, 1 or 2, we want to map 2 to 0 and the others to 1
            expr = nmlop.VACT2_CMP(expr.expr1, expr.expr2)
            # swap 0 and 2
            expr = nmlop.XOR(expr, 2)
            # map 1/2 to 1
            expr = nmlop.MIN(expr, 1)
        elif expr.op == nmlop.CMP_GE:
            # return value is 0, 1 or 2, we want to map 1/2 to 1
            expr = nmlop.VACT2_CMP(expr.expr1, expr.expr2)
            expr = nmlop.MIN(expr, 1)
        elif expr.op == nmlop.CMP_EQ:
            # return value is 0, 1 or 2, we want to map 1 to 1, other to 0
            expr = nmlop.VACT2_CMP(expr.expr1, expr.expr2)
            expr = nmlop.AND(expr, 1)
        elif expr.op == nmlop.CMP_NEQ:
            # same as CMP_EQ but invert the result
            expr = nmlop.VACT2_CMP(expr.expr1, expr.expr2)
            expr = nmlop.AND(expr, 1)
            expr = nmlop.XOR(expr, 1)

        elif expr.op == nmlop.HASBIT:
            # hasbit(x, n) ==> (x >> n) & 1
            expr = nmlop.SHIFTU_RIGHT(expr.expr1, expr.expr2)
            expr = nmlop.AND(expr, 1)
        elif expr.op == nmlop.NOTHASBIT:
            # !hasbit(x, n) ==> ((x >> n) & 1) ^ 1
            expr = nmlop.SHIFTU_RIGHT(expr.expr1, expr.expr2)
            expr = nmlop.AND(expr, 1)
            expr = nmlop.XOR(expr, 1)

        return expr.reduce()

    def preprocess_ternaryop(self, expr):
        assert isinstance(expr, expression.TernaryOp)
        if not expr.expr1.is_read_only() or not expr.expr2.is_read_only():
            return create_ternary_action(
                expr.guard,
                expr.expr1,
                expr.expr2,
                self.extra_actions,
                self.action_feature,
                self.var_scope,
                self.var_range,
            )
        guard = expression.Boolean(expr.guard).reduce()
        self.parse(guard)
        if isinstance(expr.expr1, expression.ConstantNumeric) and isinstance(expr.expr2, expression.ConstantNumeric):
            # This can be done more efficiently as (guard)*(expr1-expr2) + expr2
            self.var_list.append(nmlop.MUL)
            diff_var = VarAction2Var(0x1A, 0, expr.expr1.value - expr.expr2.value)
            diff_var.comment = "expr1 - expr2"
            self.var_list.append(diff_var)
            self.var_list.append(nmlop.ADD)
            # Add var sizes, +2 for the operators
            self.var_list_size += 2 + diff_var.get_size()
            return expr.expr2
        else:
            guard_var = VarAction2StoreTempVar()
            guard_var.comment = "guard"
            inverted_guard_var = VarAction2StoreTempVar()
            inverted_guard_var.comment = "!guard"
            self.var_list.append(nmlop.STO_TMP)
            self.var_list.append(guard_var)
            self.var_list.append(nmlop.XOR)
            var = VarAction2Var(0x1A, 0, 1)
            self.var_list.append(var)
            self.var_list.append(nmlop.STO_TMP)
            self.var_list.append(inverted_guard_var)
            self.var_list.append(nmlop.VAL2)
            # the +4 is for the 4 operators added above (STO_TMP, XOR, STO_TMP, VAL2)
            self.var_list_size += 4 + guard_var.get_size() + inverted_guard_var.get_size() + var.get_size()
            expr1 = nmlop.MUL(expr.expr1, VarAction2LoadTempVar(guard_var))
            expr2 = nmlop.MUL(expr.expr2, VarAction2LoadTempVar(inverted_guard_var))
            return nmlop.ADD(expr1, expr2)

    def preprocess_absop(self, expr):
        assert isinstance(expr, expression.AbsOp)
        self.parse(expr.expr.reduce())
        input_var = VarAction2StoreTempVar()
        input_var.comment = "input"
        self.var_list.append(nmlop.STO_TMP)
        self.var_list.append(input_var)
        self.var_list.append(nmlop.VACT2_CMP)
        var = VarAction2Var(0x1A, 0, 0)
        self.var_list.append(var)
        self.var_list.append(nmlop.SUB)
        var2 = VarAction2Var(0x1A, 0, 1)
        self.var_list.append(var2)
        self.var_list.append(nmlop.MUL)
        # the +4 is for the 4 operators added above (STO_TMP, VACT2_CMP, SUB, MUL)
        self.var_list_size += 4 + input_var.get_size() + var.get_size() + var2.get_size()
        return VarAction2LoadTempVar(input_var)

    def preprocess_storageop(self, expr):
        assert isinstance(expr, expression.StorageOp)
        if expr.info["perm"] and not self.var_scope.has_persistent_storage:
            raise generic.ScriptError(
                "Persistent storage is not supported for feature '{}'".format(self.var_scope.name),
                expr.pos,
            )

        if expr.info["store"]:
            op = nmlop.STO_PERM if expr.info["perm"] else nmlop.STO_TMP
            ret = expression.BinOp(op, expr.value, expr.register, expr.pos)
        else:
            var_num = 0x7C if expr.info["perm"] else 0x7D
            ret = expression.Variable(expression.ConstantNumeric(var_num), param=expr.register, pos=expr.pos)

        if expr.info["perm"] and self.var_scope is action2var_variables.scope_towns:
            # store grfid in register 0x100 for town persistent storage
            grfid = expression.ConstantNumeric(
                0xFFFFFFFF if expr.grfid is None else expression.parse_string_to_dword(expr.grfid)
            )
            store_op = nmlop.STO_TMP(grfid, 0x100, expr.pos)
            ret = nmlop.VAL2(store_op, ret)
        elif expr.grfid is not None:
            raise generic.ScriptError("Specifying a grfid is only possible for town persistent storage.", expr.pos)
        return ret

    def parse_expr_to_constant(self, expr, offset):
        if isinstance(expr, expression.ConstantNumeric):
            return expr.value

        tmp_param, tmp_param_actions = actionD.get_tmp_parameter(expr)
        self.extra_actions.extend(tmp_param_actions)
        self.mods.append(Modification(tmp_param, 4, self.var_list_size + offset))
        return 0

    def parse_variable(self, expr):
        """
        Parse a variable in an expression.

        @param expr:
        @type  expr: L{expression.Variable}
        """
        if not isinstance(expr.num, expression.ConstantNumeric):
            raise generic.ScriptError("Variable number must be a constant number", expr.pos)
        if not (expr.param is None or isinstance(expr.param, expression.ConstantNumeric)):
            raise generic.ScriptError("Variable parameter must be a constant number", expr.pos)

        if len(expr.extra_params) > 0:
            first_var = len(self.var_list) == 0
            backup_op = None
            value_backup = None
            if not first_var:
                backup_op = self.var_list.pop()
                value_backup = VarAction2StoreTempVar()
                self.var_list.append(nmlop.STO_TMP)
                self.var_list.append(value_backup)
                self.var_list.append(nmlop.VAL2)
                self.var_list_size += value_backup.get_size() + 1

            # Last value == 0, and this is right before we're going to use
            # the extra parameters. Set them to their correct value here.
            for extra_param in expr.extra_params:
                self.parse(extra_param[1])
                self.var_list.append(nmlop.STO_TMP)
                var = VarAction2Var(0x1A, 0, extra_param[0])
                self.var_list.append(var)
                self.var_list.append(nmlop.VAL2)
                self.var_list_size += var.get_size() + 2

            if not first_var:
                value_loadback = VarAction2LoadTempVar(value_backup)
                self.var_list.append(value_loadback)
                self.var_list.append(backup_op)
                self.var_list_size += value_loadback.get_size() + 1

        if expr.param is None:
            offset = 2
            param = None
        else:
            offset = 3
            param = expr.param.value
        mask = self.parse_expr_to_constant(expr.mask, offset)

        var = VarAction2Var(expr.num.value, expr.shift.value, mask, param)

        if expr.add is not None:
            var.add = self.parse_expr_to_constant(expr.add, offset + 4)
        if expr.div is not None:
            var.div = self.parse_expr_to_constant(expr.div, offset + 8)
        if expr.mod is not None:
            var.mod = self.parse_expr_to_constant(expr.mod, offset + 8)
        self.var_list.append(var)
        self.var_list_size += var.get_size()

    def parse_not(self, expr):
        self.parse_binop(nmlop.XOR(expr.expr, 1))

    def parse_binop(self, expr):
        if expr.op.act2_num is None:
            expr.supported_by_action2(True)

        if expr.op == nmlop.STO_TMP:
            self.in_store_op = True

        if (
            isinstance(expr.expr2, (expression.ConstantNumeric, expression.Variable))
            or isinstance(expr.expr2, (VarAction2LoadTempVar, VarAction2LoadCallParam))
            or (isinstance(expr.expr2, expression.Parameter) and isinstance(expr.expr2.num, expression.ConstantNumeric))
            or (isinstance(expr.expr2, expression.StorageOp) and expr.expr2.name == "LOAD_TEMP")
            or expr.op == nmlop.VAL2
        ):
            expr2 = expr.expr2
        elif expr.expr2.supported_by_actionD(False):
            tmp_param, tmp_param_actions = actionD.get_tmp_parameter(expr.expr2)
            self.extra_actions.extend(tmp_param_actions)
            expr2 = expression.Parameter(expression.ConstantNumeric(tmp_param))
        else:
            # The expression is so complex we need to compute it first, store the
            # result and load it back later.
            self.parse(expr.expr2)
            tmp_var = VarAction2StoreTempVar()
            self.var_list.append(nmlop.STO_TMP)
            self.var_list.append(tmp_var)
            self.var_list.append(nmlop.VAL2)
            # the +2 is for both operators
            self.var_list_size += tmp_var.get_size() + 2
            expr2 = VarAction2LoadTempVar(tmp_var)

        # parse expr1
        self.parse(expr.expr1)
        self.var_list.append(expr.op)
        self.var_list_size += 1

        self.parse(expr2)

        if expr.op == nmlop.STO_TMP:
            self.in_store_op = False

    def parse_constant(self, expr):
        var = VarAction2Var(0x1A, 0, expr.value)
        self.var_list.append(var)
        self.var_list_size += var.get_size()

    def parse_param(self, expr):
        self.mods.append(Modification(expr.num.value, 4, self.var_list_size + 2))
        var = VarAction2Var(0x1A, 0, 0)
        var.comment = str(expr)
        self.var_list.append(var)
        self.var_list_size += var.get_size()

    def parse_string(self, expr):
        str_id, actions = action4.get_string_action4s(0, 0xDC if self.in_store_op else 0xD0, expr)
        self.extra_actions.extend(actions)
        self.parse_constant(expression.ConstantNumeric(str_id))

    def parse_via_actionD(self, expr):
        tmp_param, tmp_param_actions = actionD.get_tmp_parameter(expr)
        self.extra_actions.extend(tmp_param_actions)
        num = expression.ConstantNumeric(tmp_param)
        self.parse(expression.Parameter(num))

    def parse_proc_call(self, expr):
        assert isinstance(expr, expression.SpriteGroupRef)
        assert not expr.param_list or not expr.act2
        var_access = VarAction2ProcCallVar(expr)
        target = action2.resolve_spritegroup(expr.name) if not expr.act2 else None
        refs = expr.collect_references()

        # Fill param registers for the call
        tmp_vars = []
        for i, param in enumerate(expr.param_list):
            if i > 0:  # No operator before first param as per advanced VarAct2 syntax
                self.var_list.append(nmlop.VAL2)
                self.var_list_size += 1
            if refs != [expr]:
                # For f(x, g(y)), x can be overwritten by y if f and g share the same param registers
                # Use temporary variables as an intermediate step
                store_tmp = VarAction2StoreTempVar()
                tmp_vars.append(
                    (
                        VarAction2LoadTempVar(store_tmp),
                        VarAction2StoreCallParam(target.register_map[self.action_feature][i]),
                    )
                )
            else:
                store_tmp = VarAction2StoreCallParam(target.register_map[self.action_feature][i])
            self.parse_expr(reduce_varaction2_expr(param, self.var_scope))
            self.var_list.append(nmlop.STO_TMP)
            self.var_list.append(store_tmp)
            self.var_list_size += store_tmp.get_size() + 1  # Add 1 for operator

        # Fill param registers with temporary variables if needed
        for src, dest in tmp_vars:
            self.var_list.append(nmlop.VAL2)
            self.var_list.append(src)
            self.var_list.append(nmlop.STO_TMP)
            self.var_list.append(dest)
            self.var_list_size += src.get_size() + dest.get_size() + 2  # Add 2 for operators

        if expr.param_list:
            self.var_list.append(nmlop.VAL2)
            self.var_list_size += 1
        self.var_list.append(var_access)
        self.var_list_size += var_access.get_size()
        self.proc_call_list.append(expr)

    def parse_expr(self, expr):
        if isinstance(expr, expression.Array):
            if len(expr.values) == 0:
                raise generic.ScriptError("An array of expressions cannot be empty", expr.pos)
            for expr2 in expr.values:
                self.parse(expr2)
                self.var_list.append(nmlop.VAL2)
                self.var_list_size += 1
            # Drop the trailing VAL2 again
            self.var_list.pop()
            self.var_list_size -= 1
        else:
            self.parse(expr)

    def parse(self, expr):
        # Preprocess the expression
        if isinstance(expr, expression.SpecialParameter):
            # do this first, since it may evaluate to a BinOp
            expr = expr.to_reading()

        if isinstance(expr, expression.BinOp):
            expr = self.preprocess_binop(expr)

        elif isinstance(expr, expression.Boolean):
            expr = nmlop.MINU(expr.expr, 1)

        elif isinstance(expr, expression.BinNot):
            expr = nmlop.XOR(expr.expr, 0xFFFFFFFF)

        elif isinstance(expr, expression.TernaryOp) and not expr.supported_by_actionD(False):
            expr = self.preprocess_ternaryop(expr)

        elif isinstance(expr, expression.AbsOp) and not expr.supported_by_actionD(False):
            expr = self.preprocess_absop(expr)

        elif isinstance(expr, expression.StorageOp):
            expr = self.preprocess_storageop(expr)

        # Try to parse the expression to a list of variables+operators
        if isinstance(expr, expression.ConstantNumeric):
            self.parse_constant(expr)

        elif isinstance(expr, expression.Parameter) and isinstance(expr.num, expression.ConstantNumeric):
            self.parse_param(expr)

        elif isinstance(expr, expression.Variable):
            self.parse_variable(expr)

        elif expr.supported_by_actionD(False):
            self.parse_via_actionD(expr)

        elif isinstance(expr, expression.BinOp):
            self.parse_binop(expr)

        elif isinstance(expr, expression.Not):
            self.parse_not(expr)

        elif isinstance(expr, expression.String):
            self.parse_string(expr)

        elif isinstance(expr, (VarAction2LoadTempVar, VarAction2LoadCallParam)):
            self.var_list.append(expr)
            self.var_list_size += expr.get_size()

        elif isinstance(expr, expression.SpriteGroupRef):
            self.parse_proc_call(expr)

        else:
            expr.supported_by_action2(True)
            raise AssertionError("supported_by_action2 should have raised the correct error already")


def parse_var(name, info, pos):
    if "replaced_by" in info:
        generic.print_warning(
            generic.Warning.DEPRECATION,
            "'{}' is deprecated, consider using '{}' instead".format(name, info["replaced_by"]),
            pos,
        )
    param = expression.ConstantNumeric(info["param"]) if "param" in info else None
    res = expression.Variable(
        expression.ConstantNumeric(info["var"]),
        expression.ConstantNumeric(info["start"]),
        expression.ConstantNumeric((1 << info["size"]) - 1),
        param,
        pos,
    )
    if "value_function" in info:
        return info["value_function"](res, info)
    return res


def parse_60x_var(name, args, pos, info):
    if "param_function" in info:
        # Special function to extract parameters if there is more than one
        param, extra_params = info["param_function"](name, args, pos, info)
    else:
        # Default function to extract parameters
        param, extra_params = action2var_variables.default_60xvar(name, args, pos, info)

    if isinstance(param, expression.ConstantNumeric) and (0 <= param.value <= 255):
        res = expression.Variable(
            expression.ConstantNumeric(info["var"]),
            expression.ConstantNumeric(info["start"]),
            expression.ConstantNumeric((1 << info["size"]) - 1),
            param,
            pos,
        )

        res.extra_params.extend(extra_params)
    else:
        # Make use of var 7B to pass non-constant parameters
        var = expression.Variable(
            expression.ConstantNumeric(0x7B),
            expression.ConstantNumeric(info["start"]),
            expression.ConstantNumeric((1 << info["size"]) - 1),
            expression.ConstantNumeric(info["var"]),
            pos,
        )

        var.extra_params.extend(extra_params)
        # Set the param in the accumulator beforehand
        res = nmlop.VAL2(param, var, pos)

    if "value_function" in info:
        res = info["value_function"](res, info)
    return res


def parse_minmax(value, unit_str, action_list, act6, offset):
    """
    Parse a min or max value in a switch block.

    @param value: Value to parse
    @type value: L{Expression}

    @param unit_str: Unit to use
    @type unit_str: C{str} or C{None}

    @param action_list: List to append any extra actions to
    @type action_list: C{list} of L{BaseAction}

    @param act6: Action6 to add any modifications to
    @type act6: L{Action6}

    @param offset: Current offset to use for action6
    @type offset: C{int}

    @return: A tuple of two values:
                - The value to use as min/max
                - Whether the resulting range may need a sanity check
    @rtype: C{tuple} of (L{ConstantNumeric} or L{SpriteGroupRef}), C{bool}
    """
    if unit_str is not None:
        raise generic.ScriptError("Using a unit is in switch-ranges is not (temporarily) not supported", value.pos)
    result, offset = actionD.write_action_value(value, action_list, act6, offset, 4)
    check_range = isinstance(value, expression.ConstantNumeric)
    return (result, offset, check_range)


return_action_id = 0


def create_ternary_action(guard, expr_true, expr_false, action_list, feature, var_scope, var_range):
    act6 = action6.Action6()

    global return_action_id
    name = "@ternary_action_{:d}".format(return_action_id)
    varaction2 = Action2Var(feature, name, guard.pos, var_range)
    return_action_id += 1

    expr = reduce_varaction2_expr(guard, var_scope)

    offset = 4  # first var

    parser = Varaction2Parser(feature, var_range)
    parser.parse_expr(expr)
    action_list.extend(parser.extra_actions)
    for mod in parser.mods:
        act6.modify_bytes(mod.param, mod.size, mod.offset + offset)
    varaction2.var_list = parser.var_list
    offset += parser.var_list_size + 1  # +1 for the byte num-ranges
    for proc in parser.proc_call_list:
        action2.add_ref(proc, varaction2, True)

    result, result_comment = parse_result(expr_false, action_list, act6, offset, varaction2, None, var_range)
    offset += 2  # size of result
    varaction2.ranges.append(
        VarAction2Range(expression.ConstantNumeric(0), expression.ConstantNumeric(0), result, result_comment)
    )

    default, default_comment = parse_result(expr_true, action_list, act6, offset, varaction2, None, var_range)
    varaction2.default_result = default
    varaction2.default_comment = default_comment

    if len(act6.modifications) > 0:
        action_list.append(act6)

    action_list.append(varaction2)
    return expression.SpriteGroupRef(expression.Identifier(name), [], None, varaction2)


def create_return_action(expr, feature, name, var_range):
    """
    Create a varaction2 to return the computed value

    @param expr: Expression to return
    @type expr: L{Expression}

    @param feature: Feature of the switch-block
    @type feature: C{int}

    @param name: Name of the new varaction2
    @type name: C{str}

    @return: A tuple of two values:
                - Action list to prepend
                - Reference to the created varaction2
    @rtype: C{tuple} of (C{list} of L{BaseAction}, L{SpriteGroupRef})
    """
    varact2parser = Varaction2Parser(feature, var_range)
    varact2parser.parse_expr(expr)

    action_list = varact2parser.extra_actions
    extra_act6 = action6.Action6()
    for mod in varact2parser.mods:
        extra_act6.modify_bytes(mod.param, mod.size, mod.offset + 4)
    if len(extra_act6.modifications) > 0:
        action_list.append(extra_act6)

    varaction2 = Action2Var(feature, name, expr.pos, var_range)
    varaction2.var_list = varact2parser.var_list
    varaction2.default_result = expression.ConstantNumeric(0)  # Bogus result, it's the nvar == 0 that matters
    varaction2.default_comment = "Return computed value"

    for proc in varact2parser.proc_call_list:
        action2.add_ref(proc, varaction2, True)

    ref = expression.SpriteGroupRef(expression.Identifier(name), [], None, varaction2)
    action_list.append(varaction2)
    return (action_list, ref)


failed_cb_results = {}


def get_failed_cb_result(feature, action_list, parent_action, pos):
    """
    Get a sprite group reference to use for a failed callback
    The actions needed are created on first use, then cached in L{failed_cb_results}

    @param feature: Feature to use
    @type feature: C{int}

    @param action_list: Action list to append any extra actions to
    @type action_list: C{list} of L{BaseAction}

    @param parent_action: Reference to the action of which this is a result
    @type parent_action: L{BaseAction}

    @param pos: Positional context.
    @type  pos: L{Position}

    @return: Sprite group reference to use
    @rtype: L{SpriteGroupRef}
    """
    if feature in failed_cb_results:
        varaction2 = failed_cb_results[feature]
    else:
        # Create action2 (+ action1, if needed)
        # Import here to avoid circular imports
        from nml.actions import action1, action2layout, action2production, action2real

        if feature == 0x0A:
            # Industries -> production action2
            act2 = action2production.make_empty_production_action2(pos)
        elif feature in (0x07, 0x09, 0x0F, 0x11, 0x14):
            # Tile layout action2
            act2 = action2layout.make_empty_layout_action2(feature, pos)
        else:
            # Normal action2
            act1_actions, act1_index = action1.make_cb_failure_action1(feature)
            action_list.extend(act1_actions)
            act2 = action2real.make_simple_real_action2(
                feature, "@CB_FAILED_REAL{:02X}".format(feature), pos, act1_index
            )
        action_list.append(act2)

        # Create varaction2, to choose between returning graphics and 0, depending on CB
        varact2parser = Varaction2Parser(feature)
        varact2parser.parse_expr(
            expression.Variable(expression.ConstantNumeric(0x0C), mask=expression.ConstantNumeric(0xFFFF))
        )

        varaction2 = Action2Var(feature, "@CB_FAILED{:02X}".format(feature), pos, 0x89)
        varaction2.var_list = varact2parser.var_list

        varaction2.ranges.append(
            VarAction2Range(
                expression.ConstantNumeric(0),
                expression.ConstantNumeric(0),
                expression.ConstantNumeric(0),
                "graphics callback -> return 0",
            )
        )
        varaction2.default_result = expression.SpriteGroupRef(expression.Identifier(act2.name), [], None, act2)
        varaction2.default_comment = "Non-graphics callback, return graphics result"
        action2.add_ref(varaction2.default_result, varaction2)

        action_list.append(varaction2)
        failed_cb_results[feature] = varaction2

    ref = expression.SpriteGroupRef(expression.Identifier(varaction2.name), [], None, varaction2)
    action2.add_ref(ref, parent_action)
    return ref


def parse_sg_ref_result(result, action_list, parent_action, var_range):
    """
    Parse a result that is a sprite group reference.

    @param result: Result to parse
    @type result: L{SpriteGroupRef}

    @param action_list: List to append any extra actions to
    @type action_list: C{list} of L{BaseAction}

    @param parent_action: Reference to the action of which this is a result
    @type parent_action: L{BaseAction}

    @param var_range: Variable range to use for variables in the expression
    @type var_range: C{int}

    @return: Result to use in the calling varaction2
    @rtype: L{SpriteGroupRef}
    """
    if result.name.value == "CB_FAILED":
        return get_failed_cb_result(parent_action.feature, action_list, parent_action, result.pos)

    target = action2.resolve_spritegroup(result.name) if not result.act2 else None

    if parent_action.feature not in action2.features_sprite_layout and isinstance(target, spriteblock.SpriteLayout):
        raise generic.ScriptError(
            "SpriteLayout '{}' is not a valid return value for feature '{}'".format(
                result.name, general.feature_name(parent_action.feature)
            ),
            result.pos,
        )

    if len(result.param_list) == 0:
        action2.add_ref(result, parent_action)
        return result

    # Result is parametrized
    # Insert an intermediate varaction2 to store expressions in registers
    var_scope = get_scope(parent_action.feature, var_range)
    varact2parser = Varaction2Parser(parent_action.feature, var_range)
    for i, param in enumerate(result.param_list):
        if i > 0:
            varact2parser.var_list.append(nmlop.VAL2)
            varact2parser.var_list_size += 1
        varact2parser.parse_expr(reduce_varaction2_expr(param, var_scope))
        varact2parser.var_list.append(nmlop.STO_TMP)
        store_tmp = VarAction2StoreCallParam(target.register_map[parent_action.feature][i])
        varact2parser.var_list.append(store_tmp)
        varact2parser.var_list_size += store_tmp.get_size() + 1  # Add 1 for operator

    action_list.extend(varact2parser.extra_actions)
    extra_act6 = action6.Action6()
    for mod in varact2parser.mods:
        extra_act6.modify_bytes(mod.param, mod.size, mod.offset + 4)
    if len(extra_act6.modifications) > 0:
        action_list.append(extra_act6)

    global return_action_id
    name = "@return_action_{:d}".format(return_action_id)
    varaction2 = Action2Var(parent_action.feature, name, result.pos, var_range)
    return_action_id += 1
    varaction2.var_list = varact2parser.var_list
    ref = expression.SpriteGroupRef(result.name, [], result.pos)
    varaction2.ranges.append(
        VarAction2Range(expression.ConstantNumeric(0), expression.ConstantNumeric(0), ref, result.name.value)
    )
    varaction2.default_result = ref
    varaction2.default_comment = result.name.value
    # Add the references as procs, to make sure, that any intermediate registers
    # are freed at the spritelayout and thus not selected to pass parameters
    # Reference is used twice (range + default) so call add_ref twice
    action2.add_ref(ref, varaction2, True)
    action2.add_ref(ref, varaction2, True)

    ref = expression.SpriteGroupRef(expression.Identifier(name), [], None, varaction2)
    action_list.append(varaction2)
    action2.add_ref(ref, parent_action)

    return ref


def parse_result(value, action_list, act6, offset, parent_action, none_result, var_range, repeat_result=1):
    """
    Parse a result (another switch or CB result) in a switch block.

    @param value: Value to parse
    @type value: L{Expression}

    @param action_list: List to append any extra actions to
    @type action_list: C{list} of L{BaseAction}

    @param act6: Action6 to add any modifications to
    @type act6: L{Action6}

    @param offset: Current offset to use for action6
    @type offset: C{int}

    @param parent_action: Reference to the action of which this is a result
    @type parent_action: L{BaseAction}

    @param none_result: Result to use to return the computed value
    @type none_result: L{Expression}

    @param var_range: Variable range to use for variables in the expression
    @type var_range: C{int}

    @param repeat_result: Repeat any action6 modifying of the next sprite this many times.
    @type repeat_result: C{int}

    @return: A tuple of two values:
                - The value to use as return value
                - Comment to add to this value
    @rtype: C{tuple} of (L{ConstantNumeric} or L{SpriteGroupRef}), C{str}
    """
    if value is None:
        comment = "return;"
        assert none_result is not None
        if isinstance(none_result, expression.SpriteGroupRef):
            result = parse_sg_ref_result(none_result, action_list, parent_action, var_range)
        else:
            result = none_result
    elif isinstance(value, expression.SpriteGroupRef):
        result = parse_sg_ref_result(value, action_list, parent_action, var_range)
        comment = result.name.value + ";"
    elif isinstance(value, expression.ConstantNumeric):
        comment = "return {:d};".format(value.value)
        result = value
        if not (-16384 <= value.value <= 32767):
            msg = (
                "Callback results are limited to -16384..16383 (when the result is a signed number)"
                " or 0..32767 (unsigned), encountered {:d}."
            ).format(value.value)
            raise generic.ScriptError(msg, value.pos)

    elif isinstance(value, expression.String):
        comment = "return {};".format(str(value))
        str_id, actions = action4.get_string_action4s(0, 0xD0, value)
        action_list.extend(actions)
        result = expression.ConstantNumeric(str_id - 0xD000 + 0x8000)
    elif value.supported_by_actionD(False):
        tmp_param, tmp_param_actions = actionD.get_tmp_parameter(nmlop.OR(value, 0x8000).reduce())
        comment = "return param[{:d}];".format(tmp_param)
        action_list.extend(tmp_param_actions)
        for i in range(repeat_result):
            act6.modify_bytes(tmp_param, 2, offset + 2 * i)
        result = expression.ConstantNumeric(0)
    else:
        global return_action_id
        extra_actions, result = create_return_action(
            value, parent_action.feature, "@return_action_{:d}".format(return_action_id), var_range
        )
        return_action_id += 1
        action2.add_ref(result, parent_action)
        action_list.extend(extra_actions)
        comment = "return {}".format(value)
    return (result, comment)


def get_scope(action_feature, var_range=0x89):
    return action2var_variables.varact2features[action_feature].get_scope(var_range)


def reduce_varaction2_expr(expr, var_scope, extra_dicts=None):
    # 'normal' and 60+x variables to use
    vars_normal = var_scope.vars_normal
    vars_60x = var_scope.vars_60x

    def func60x(name, value, pos):
        return expression.FunctionPtr(expression.Identifier(name, pos), parse_60x_var, value)

    id_dicts = (
        (extra_dicts or [])
        + [(action2var_variables.varact2_globalvars, parse_var), (vars_normal, parse_var), (vars_60x, func60x)]
        + global_constants.const_list
    )

    return expr.reduce(id_dicts)


def parse_varaction2(switch_block):
    global return_action_id
    return_action_id = 0

    action6.free_parameters.save()
    act6 = action6.Action6()
    action_list = action2real.create_spriteset_actions(switch_block)

    feature = next(iter(switch_block.feature_set))
    var_scope = get_scope(feature, switch_block.var_range)
    varaction2 = Action2Var(
        feature,
        switch_block.name.value,
        switch_block.pos,
        switch_block.var_range,
        switch_block.register_map[feature],
    )

    expr = reduce_varaction2_expr(switch_block.expr, var_scope)

    offset = 4  # first var

    parser = Varaction2Parser(feature, switch_block.var_range)
    parser.parse_expr(expr)
    action_list.extend(parser.extra_actions)
    for mod in parser.mods:
        act6.modify_bytes(mod.param, mod.size, mod.offset + offset)
    varaction2.var_list = parser.var_list
    offset += parser.var_list_size + 1  # +1 for the byte num-ranges
    for proc in parser.proc_call_list:
        action2.add_ref(proc, varaction2, True)

    none_result = None
    if any(
        x is not None and x.value is None
        for x in [r.result for r in switch_block.body.ranges] + [switch_block.body.default]
    ):
        # Computed result is returned in at least one result
        if len(switch_block.body.ranges) == 0:
            # There is only a default, which is 'return computed result', so we're fine
            none_result = expression.ConstantNumeric(0)  # Return value does not matter
        else:
            # Add an extra action to return the computed value
            extra_actions, none_result = create_return_action(
                expression.Variable(expression.ConstantNumeric(0x1C)),
                feature,
                switch_block.name.value + "@return",
                0x89,
            )
            action_list.extend(extra_actions)

    used_ranges = []
    for r in switch_block.body.ranges:
        comment = str(r.min) + " .. " + str(r.max) + ": "

        range_result, range_comment = parse_result(
            r.result.value, action_list, act6, offset, varaction2, none_result, switch_block.var_range
        )
        comment += range_comment
        offset += 2  # size of result

        range_min, offset, check_min = parse_minmax(r.min, r.unit, action_list, act6, offset)
        range_max, offset, check_max = parse_minmax(r.max, r.unit, action_list, act6, offset)

        range_overlap = False
        if check_min and check_max:
            for existing_range in used_ranges:
                if existing_range[0] <= range_min.value and range_max.value <= existing_range[1]:
                    generic.print_warning(
                        generic.Warning.GENERIC,
                        "Range overlaps with existing ranges so it'll never be reached",
                        r.min.pos,
                    )
                    range_overlap = True
                    break
            if not range_overlap:
                used_ranges.append([range_min.value, range_max.value])
                used_ranges.sort()
                i = 0
                while i + 1 < len(used_ranges):
                    if used_ranges[i + 1][0] <= used_ranges[i][1] + 1:
                        used_ranges[i][1] = max(used_ranges[i][1], used_ranges[i + 1][1])
                        used_ranges.pop(i + 1)
                    else:
                        i += 1

        if not range_overlap:
            varaction2.ranges.append(VarAction2Range(range_min, range_max, range_result, comment))

    if len(switch_block.body.ranges) == 0 and (
        switch_block.body.default is None or switch_block.body.default.value is not None
    ):
        # Computed result is not returned, but there are no ranges
        # Add one range, to avoid the nvar == 0 bear trap
        offset += 10
        range_result, range_comment = parse_result(
            expression.SpriteGroupRef(expression.Identifier("CB_FAILED", None), [], None),
            action_list,
            act6,
            offset,
            varaction2,
            none_result,
            switch_block.var_range,
        )
        varaction2.ranges.append(
            VarAction2Range(
                expression.ConstantNumeric(1),
                expression.ConstantNumeric(0),
                range_result,
                "Bogus range to avoid nvar == 0",
            )
        )

    if len(varaction2.ranges) > 255:
        raise generic.ScriptError("Too many ranges.", varaction2.pos)

    # Handle default result
    if switch_block.body.default is not None:
        # there is a default value
        default_result = switch_block.body.default.value
    else:
        # Default to CB_FAILED
        default_result = expression.SpriteGroupRef(expression.Identifier("CB_FAILED", None), [], None)

    default, default_comment = parse_result(
        default_result, action_list, act6, offset, varaction2, none_result, switch_block.var_range
    )
    varaction2.default_result = default
    if switch_block.body.default is None:
        varaction2.default_comment = "No default specified -> fail callback"
    elif switch_block.body.default.value is None:
        varaction2.default_comment = "Return computed value"
    else:
        varaction2.default_comment = "default: " + default_comment

    if len(act6.modifications) > 0:
        action_list.append(act6)

    action_list.append(varaction2)
    switch_block.set_action2(varaction2, feature)

    action6.free_parameters.restore()
    return action_list
